#!/usr/bin/env perl
# vdt 3.0.0 release
#################################################################
#
# Purpose: Perform grid site verification checks on remote
#          host(s) and report results.
#
# Author:  Craig Prescott
#
# TODO/WISHLIST:
#  * colorize responses? (green=good, red=bad, uncolored=neutral)
#  * reduce globus-job-run calls (somewhat done via caching)
#
#################################################################

#
# This script depends upon a few CPAN modules that may not
# be included in your perl distribution,  notably:
#
# Net::LDAP, Net::LDAP::Filter
#
# If they are not, and you do not wish to install them 
# globally (or are not able to due to privileges), you can 
# still run this script by 
# 
# 1) installing them wherever you have write access
# 2) and, EITHER
#    * Add that directory to your PERL5LIB environment variable.
#    OR
#    * Uncomment the BEGIN block below.
#      Replace "YOUR_PERSONAL_PERLDIR" with the path to your
#      module installation(s).  You may add as many "push"
#      statements as you like to the BEGIN block.
#
# You may find the following web page of some use:
#
#   http://www.phys.ufl.edu/~prescott/perl_modules_normal_user.html
#
#BEGIN { push(@INC, "YOUR_PERSONAL_PERLDIR"); }

#######################################################################
#
# $tests->{STATUS}{$testname} = SUCCESS, FAILURE, or UNTESTED
# $tests->{STDOUT}{$testname} = text output to be display
# $tests->{STDERR}{$testname} = text output to be display
#
# $opts->{TESTNAME}                  name of current test
#
# $opts->{FLAGS}{VERBOSE}            verbosity flag
# $opts->{FLAGS}{DEBUG}              debug flag
#
# $opts->{GATEKEEPER_HOST}           hostname of remote gatekeeper service
# $opts->{GATEKEEPER_PORT}           port number of remote gatekeeper service
# $opts->{GSIFTP_HOST}               hostname of remote gsiftp service
# $opts->{GSIFTP_PORT}               port number of remote gsiftp service
# $opts->{MDS_HOST}                  hostname of remote MDS service
# $opts->{MDS_PORT}                  port number of remote MDS service
#
# The $opts->{LOCAL_FILES} hash has the following keys:
#
#   SERVICES            Local filename of remote /etc/services
#   INETD_CONF          Local filename of remote /etc/inetd.conf or xinetd.conf,
#                       as appropriate.  For xinetd.conf, any includedir 
#                       directives are processed and substituted.
#   OSG_ATTRIBUTES_CONF Local filename of remote osg-attributes.conf
#
# The $opts->{REMOTE_FILES} hash has the following keys:
#
#   INETD_CONF      Path to remote inet super-server configuration file.
#
# The $opts->{FILE_HASH} hash has the following keys:
#
#
# $opts->{FILE_HASH}{INETD_CONF}     Representation of the remote site's
#                                    inet super-server configuration file
#                                    (/etc/inetd.conf or /etc/xinetd.conf).
#                                    For xinetd.conf, any includedir 
#                                    directives are processed and substituted.
#                                    The hash contains the following keys:
#
#                                    {service_name}{protocol}{socket_type}
#                                    {service_name}{protocol}{wait}
#                                    {service_name}{protocol}{user}
#                                    {service_name}{protocol}{server}
#                                    {service_name}{protocol}{server_args}
#
#                                    Additional keys exist for each attribute
#                                    in xinetd configurations.
#
# $opts->{FILE_HASH}{SERVICES}       Representation of the remote site's
#                                    Internet network services list.
#                                    The hash contains the following keys:
#
#                                    {service_name}{protocol}{PORT}
#                                    {service_name}{protocol}{ALIASES}
#
# $opts->{JM_TYPES}{$jm} = $type;
# $opts->{SCHED_PATHS}{$type} = $path;
#
#################################################################

use Getopt::Long;
use File::Basename;
use Sys::Hostname;
use Socket;
use IO::File;
use IO::Socket::INET;
#use Net::LDAP;
#use Net::LDAP::Filter;
use File::Temp qw(tempfile);
use File::Compare;

$| = 1;

#
# Define some constants
#

# STATUS constants
use constant UNTESTED => 'UNTESTED';
use constant SUCCESS  => 'PASS';
use constant FAILURE  => 'FAIL';

# Other constants
use constant UNKNOWN  => 'UNKNOWN';
use constant YES      => 'YES';
use constant NO       => 'NO';

#
# get_fqdn
#
# Return the fully qualified domain name associated with the input.
#
sub get_fqdn($)
{
    my $in = shift;

    if ( ! defined $in ) { return undef; }

    my $out = $in;

    if ( lc($in) eq 'localhost' )
    {
	$out = hostname();
	my $address = gethostbyname($out)
	    or die "Couldn't resolve $out: $!";
	$out = gethostbyaddr($address, AF_INET)
	    or die "Couldn't re-resolve $out: $!";
    }

    return $out;
}


#
# stdout_to_file
#
# Dump a string to a file, splitting on newlines, and return the filename.
#
sub stdout_to_file($$)
{
    my ($s, $opts) = @_;

    my ($fh, $f) = tempfile(
			    TEMPLATE => 'remoteXXXXXX',
			    DIR      => $opts->{'TMPDIR'}, 
			    UNLINK   => 0,
			    );

    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $f created\n"; }

    my @lines = split "\n", $s;

    foreach my $line (@lines)
    {
	print $fh "$line\n";
    }

    close $fh;

    return $f;
}


#
# get_service_name
#
# Return the name of the service running on the supplied port with the
# supplied protocol from the remote site's Internet network services list.
#
sub get_service_name($$$)
{
    my ($opts, $port, $prot) = @_;

    if ( ! defined $opts ) { return undef; }
    if ( ! defined $port ) { return undef; }
    if ( ! defined $prot ) { $prot = 'tcp'; }

    if ( ! exists $opts->{'FILE_HASH'}{'SERVICES'} ) { return undef; }

    my $svcname = undef;

    # XXX could be faster
    foreach my $svc ( keys %{$opts->{'FILE_HASH'}{'SERVICES'}} )
    {
	if ( exists $opts->{'FILE_HASH'}{'SERVICES'}{$svc}{$prot}{'PORT'} &&
	     ($opts->{'FILE_HASH'}{'SERVICES'}{$svc}{$prot}{'PORT'} == $port) )
	{
	    $svcname = $svc;
	    last;
	}
    }

    return $svcname;
}


#
# get_inetd_service_config
#
# Return a hash reference containing the remote site's inetd/xinetd 
# config for the service name and protocol supplied.
#
sub get_inetd_service_config($$$)
{
    my ($opts, $svcname, $prot) = @_;

    if ( ! defined $opts ) { return undef; }
    if ( ! defined $svcname ) { return undef; }
    if ( ! defined $prot ) { $prot = 'tcp'; }

    if ( ! exists  $opts->{'FILE_HASH'}{'INETD_CONF'} ) { return undef; }

    if ( exists $opts->{'FILE_HASH'}{'INETD_CONF'}{$svcname}{$prot} )
    {
	return $opts->{'FILE_HASH'}{'INETD_CONF'}{$svcname}{$prot};
    }

    return undef;
}


#
# run_remote_cmds
#
# Run commands at the remote site using its gatekeeper service.
# Return value is a hash reference containing keys:
#
# Cmd              - The command run
# ExitStatus       - Exit status of globus-job-run
# RemoteExitStatus - Exit status of command executed remotely
# Stdout           - Standard output of remote command
# Stderr           - Standard error of remote command
#
sub run_remote_cmds($$)
{
    my ($cmds, $ctx) = @_;

    my @a;

    if ( ! defined $cmds || scalar(@$cmds) == 0 ) { return \@a; }

    my $host = undef;
    if ( exists $ctx->{'GATEKEEPER_HOST'} )
    {
	$host = $ctx->{'GATEKEEPER_HOST'};
    }

    if ( ! defined $host ) { return \@a; }

    my $port = 2119;
    if ( exists $ctx->{'GATEKEEPER_PORT'} )
    {
	$port = $ctx->{'GATEKEEPER_PORT'};
    }

    my $dest = $host;
    if ( defined $port ) { $dest .= ":$port" }

    # Note: To avoid having to authenticate once for each supplied
    # command, which is very expensive, we glom all the commands 
    # together and sort it out later.

    my ($outfh, $outf) = tempfile(
				  TEMPLATE => 'outXXXXXX',
				  DIR      => $ctx->{'TMPDIR'},
				  UNLINK   => 0,
				  );

    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $outf created\n"; }
    my ($errfh, $errf) = tempfile(
				  TEMPLATE => 'errXXXXXX',
				  DIR      => $ctx->{'TMPDIR'},
				  UNLINK   => 0,			       
				  );
    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $errf created\n"; }

    my $c = "globus-job-run $dest/jobmanager-fork /bin/sh -c \"";

    foreach my $cmd (@$cmds)
    {
	my $h = {};

	$h->{'Cmd'}              = undef;
	$h->{'ExitStatus'}       = undef;
	$h->{'RemoteExitStatus'} = undef;
	$h->{'Stdout'}           = undef;
	$h->{'Stderr'}           = undef;


	$h->{'Cmd'} = "globus-job-run $dest/jobmanager-fork /bin/sh -c \"$cmd\"";

	$c .= "$cmd; echo RemoteExitStatus \\\$?; echo MARK 1>&2; ";

	push @a, $h;
    }

    $c .= "\" >> $outf 2>> $errf";

    my $rc = system($c);

    foreach my $i (0 .. $#a)
    {
	$a[$i]{'ExitStatus'} = $rc;
    }

    $outfh = new IO::File;

    if ( $outfh->open("< $outf") )
    {
	# Remote command exit status lives in stdout file; sort that out
	my $testnum = 0;
	my @stdout;

	while ( my $line = $outfh->getline )
	{
	    chomp($line);
	    if ( $line =~ /^RemoteExitStatus/ )
	    {
		$a[$testnum]{'Stdout'} = join "\n", @stdout;
		($a[$testnum]{'RemoteExitStatus'} = $line) =~ 
		    s/^RemoteExitStatus\s+//;
		splice @stdout;
		$testnum++;
	    }
	    else
	    {
		push @stdout, $line;
	    }
	}
    }

    unlink $outf;
    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $outf\n"; }

    $errfh = new IO::File;

    if ( $errfh->open("< $errf") )
    {
	my $testnum = 0;
	my @stderr;

	while ( my $line = $errfh->getline )
	{
	    chomp($line);
	    if ( $line =~ /^MARK/ )
	    {
		$a[$testnum]{'Stderr'} = join "\n", @stderr;
		splice @stderr;
		$testnum++;
	    }
	    else
	    {
		push @stderr, $line;
	    }
	}
    }

    unlink $errf;
    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $errf\n"; }

    return \@a;
}


#
# run_cmds
#
# Run commands on the localhost.
# Return value is a hash reference containing keys:
#
# Cmd              - The command run
# ExitStatus       - Exit status of command
# Stdout           - Standard output of command
# Stderr           - Standard error of command
#
sub run_cmds($$)
{
    my ($cmds, $ctx) = @_;

    my @a;

    if ( ! defined $cmds || scalar(@$cmds) == 0 ) { return \@a; }

    # Running commands sequentially

    foreach my $c (@$cmds)
    {
	my ($outfh, $outf) = tempfile(
				      TEMPLATE => 'outXXXXXX',
				      DIR      => $ctx->{'TMPDIR'},
				      UNLINK   => 0,
				      );
	if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $outf created\n"; }

	my ($errfh, $errf) = tempfile(
				      TEMPLATE => 'errXXXXXX',
				      DIR      => $ctx->{'TMPDIR'},
				      UNLINK   => 0,			       
				      );
	if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $errf created\n"; }

	$c .= " >> $outf 2>> $errf";
	
	my $h = {};

	$h->{'Cmd'}        = $c;
	$h->{'ExitStatus'} = system($c);
	$h->{'Stdout'}     = undef;
	$h->{'Stderr'}     = undef;

	my @lines;

	$outfh = new IO::File;

	if ( $outfh->open("< $outf") )
	{
	    while ( my $line = $outfh->getline )
	    {
		chomp($line);
		push @lines, $line;
	    }

	    $h->{'Stdout'} = join "\n", @lines;
	}

	unlink $outf;
	if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $outf\n"; }

	$errfh = new IO::File;

	if ( $errfh->open("< $errf") )
	{
	    while ( my $line = $errfh->getline )
	    {
		chomp($line);
		push @lines, $line;
	    }

	    $h->{'Stderr'} = join "\n", @lines;
	}

	unlink $errf;
	if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $errf\n"; }

	push @a, $h;
    }

    return \@a;
}


#
# hashify_xinetd_services
#
# Return a hash reference representing xinetd-style services contained
# in the given file.
#
sub hashify_xinetd_services($)
{
    my $f = shift;

    if ( ! defined $f ) { return undef; }

    my %h;

    my $fh = new IO::File;

    if ( $fh->open("< $f") )
    {
	my $svcname = undef;

	while ( my $line = $fh->getline )
	{
	    chomp($line);
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if    ( ! defined $line || length($line) == 0 ) { next;}
	    elsif ( $line =~ m/^\#/ ) { next; }
	    elsif ( $line =~ m/defaults\s*/ )
	    {
		$svcname = 'defaults';
		next;
	    }
	    elsif ( $line =~ m/\}/ )
	    {
		$svcname = undef;
		next;
	    }
	    elsif ( $line =~ m/^service\s+.*$/ )
	    {
		$svcname = $line;
		$svcname =~ s/^service\s+(\S+)/$1/;
		$h{$svcname} = undef;
		next;
	    }
	    elsif ( $line =~ m/\{/ )
	    {
		my %svctmp;

		while ( $line !~ m/^\}/ )
		{
		    $line = $fh->getline;
		    chomp($line);
		    $line =~ s/^\s+//;
		    $line =~ s/\s+$//;

		    if    ( ! defined $line || length($line) == 0 ) { next;}
		    elsif ( $line =~ m/^\#/ ) { next; }
		    elsif ( $line =~ m/^\}/ ) { last; }

		    my $key = $line;
		    $key =~ s/^(\S+)\s+.*/$1/;
		    my $op  = $line;
		    $op  =~ s/^\S+\s+(\S+)\s+.*/$1/;
		    my $val = $line;
		    $val =~ s/^\S+\s+\S+\s+(.*)$/$1/;

		    # XXX FIXME
		    # Ew - this is icky
		    # If the server arg is a shell script, it might
		    # contain useful stuff inside.  See if we can get 
		    # at it.
		    if ( $key eq 'server' && $val =~ /\.sh$/ )
		    {
			# For now, just get what would be exec'd
			my @cmds;
			push @cmds, "cat $val | grep -e \"^exec\" | tail -1 | sed 's/^exec\ +//'";
			my $a = run_remote_cmds(\@cmds, $opts);
			if ( scalar(@$a) != 0 )
			{
			    if ( $a->[0]{'ExitStatus'} == 0 && 
				 $a->[0]{'RemoteExitStatus'} == 0 )
			    {
				my @foo;
				@foo = split /\s+/, $a->[0]{'Stdout'};
				if ( scalar(@foo) > 0 )
				{
				    $val = $foo[0];
				    shift @foo;
				}
				if ( scalar(@foo) > 0 )
				{
				    # these must be args...
				    if ( ! exists $svctmp{'server_args'} )
				    {
					$svctmp{'server_args'} = 
					    join ' ', @foo;
				    }
				    else
				    {
					$svctmp{'server_args'} .= ' ' . 
					    join ' ', @foo;
				    }
				}
			    }
			}
		    }

		    if ( $op eq '=' )
		    {
			$svctmp{$key} = $val;
		    }
		    elsif ( $op eq '+=' )
		    {
			$svctmp{$key} .= ' ' . $val;
		    }
		}

		my $protocol = 'tcp';
		if ( exists $svctmp{'protocol'} )
		{
		    $protocol = $svctmp{'protocol'};
		    delete $svctmp{'protocol'};
		}

		foreach my $key ( keys %svctmp )
		{
		    $h{$svcname}{$protocol}{$key} = $svctmp{$key};
		}
	    }
	}
    }

    # Copy the defaults into other service descriptions
    # XXX 'tcp' for 'defaults' is a hack - FIXME XXX
    if ( exists $h{'defaults'}{'tcp'} )
    {
	foreach my $key ( keys %{$h{'defaults'}{'tcp'}} )
	{
	    foreach my $svc (keys %h)
	    {
		if ( $svc eq 'defaults' ) { next; }

		foreach my $prot (keys %{$h{$svc}})
		{
		    if ( ! defined $h{$svc}{$prot}{$key} )
		    {
			$h{$svc}{$prot}{$key} = $h{'defaults'}{'tcp'}{$key};
		    }
		}
	    }
	}

	undef $h{'defaults'};
    }

    return \%h;
}


#
# hashify_inetd_services
#
# Return a hash reference representing inetd-style services contained
# in the given file.
#
sub hashify_inetd_services($)
{
    my $f = shift;

    if ( ! defined $f ) { return undef; }

    my %h;

    my $fh = new IO::File;

    if ( $fh->open("< $f") )
    {
	while ( my $line = $fh->getline )
	{
	    chomp($line);
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if    ( ! defined $line || length($line) == 0 ) { next;}
	    elsif ( $line =~ m/^\#/ ) { next; }
	}
    }

    return \%h;
}


#
# hashify_inet_conf
#
# Return a hash representing the given Internet Services Daemon 
# configuration file.
#
sub hashify_inet_conf($$)
{
    my ($f_style, $f_local) = @_;

    if ( ! defined $f_style ) { return; }
    if ( ! defined $f_local )  { return; }

    my ($base, $path) = fileparse($f_style);

    my $href = undef;
    if ( $base eq "xinetd.conf" )
    {
	$href = hashify_xinetd_services($f_local);
    }
    elsif ( $base eq "inetd.conf" )
    {
	$href = hashify_inetd_services($f_local);
    }

    return $href;
}


#
# hashify_grid_mapfile
#
# Return a hash reference representing the contents of a grid-mapfile.
#
# %hash{$remote_user}->{@DNlist}
#
sub hashify_grid_mapfile($)
{
    my $f = shift;

    if ( ! defined $f || ! -f $f ) { return undef; }

    open(GMAP, "< $f");

    my %h;

    while ( my $line = <GMAP> )
    {
	chomp($line);
	if ( ! defined $line || length($line) == 0 ) { next; }
	if ( $line =~ m/^\#/ ) { next; }

	$line =~ s/^\s+//;
	$line =~ s/\s+$//;

	my $user = $line;
	$user =~ s/^.*\s+(\S+)$/$1/;
	my $dn = $line;
	$dn =~ s/\s+$user$//;
	$dn =~ s/^\"//;
	$dn =~ s/\"$//;

	$h{$dn} = $user;
    }

    close(GMAP);

    return \%h;
}


#
# obtain_remote_file
#
# Get remote file to localhost; return local filename.
#
sub obtain_remote_file($$$)
{
    my ($tests, $opts, $rfile)  = @_;

    if ( ! defined $tests ) { return undef; }
    if ( ! defined $rfile ) { return undef; }

    # Prefer gsiftp
    my $use_gsiftp = undef;

    # determine if gsiftp was reliable or not
    if ( $tests->{'STATUS'}{'gsiftp_service'}     eq SUCCESS ||
	 $tests->{'STATUS'}{'gsiftp_local_out'}   eq SUCCESS ||
	 $tests->{'STATUS'}{'gsiftp_local_in'}    eq SUCCESS ||
	 $tests->{'STATUS'}{'gsiftp_local_diffs'} eq SUCCESS )
    {
	$use_gsiftp = 1;
    }
	
    my $host = undef;
    my $port = 2811;

    my $lfile = undef;

    my ($l, $ltmp) = tempfile(
			      TEMPLATE => 'remoteXXXXXX',
			      DIR      => $opts->{'TMPDIR'}, 
			      UNLINK   => 0,
			      );
    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $ltmp created\n"; }
    if ( defined $use_gsiftp )
    {
	if ( exists $opts->{'GSIFTP_HOST'} )
	{
	    $host = $opts->{'GSIFTP_HOST'};
	}
	elsif ( exists $opts->{'GATEKEEPER_HOST'} )
	{
	    $host = $opts->{'GATEKEEPER_HOST'};
	}

	if ( exists $opts->{'GSIFTP_PORT'} )
	{
	    $port = $opts->{'GSIFTP_PORT'};
	}
	if ( defined $host )
	{
	    my $remote = $host;
	    if ( defined $port ) { $remote .= ":$port" }

	    my @cmds;
            push @cmds, "globus-url-copy gsiftp://$remote/$rfile file:$ltmp";
	    my $a = run_cmds(\@cmds, $opts);

	    if ( $a->[0]{'ExitStatus'} == 0 ) { $lfile = $ltmp }
	}
    }

    if ( ! defined $lfile )
    {
	my @cmds;
	push @cmds, "cat $rfile";
	my $a = run_remote_cmds(\@cmds, $opts);

	if ( scalar(@$a) != 0 )
	{
	    if ( $a->[0]{'ExitStatus'} == 0 && 
		 $a->[0]{'RemoteExitStatus'} == 0 )
	    {
		my @f = split "\n", $a->[0]{'Stdout'};

		if ( scalar(@f) > 0 )
		{
		    for ( my $i = 0 ; $i < scalar(@f) ; $i++ )
		    {
			print $l "$f[$i]\n";
		    }
		}

		$lfile = $ltmp
	    }
	}
    }

    return $lfile;
}


#
# hashify_remote_inet_conf
#
# Return a hash reference representing the remote site's 
# Internet Services Daemon configuration file.
#
# Prefer xinetd to inetd.
#
# Fills: $opts->{LOCAL_FILES}{INETD_CONF}
#        $opts->{REMOTE_FILES}{INETD_CONF}
#        $opts->{FILE_HASH}{INETD_CONF}
#
sub hashify_remote_inet_conf($$)
{
    my ($tests, $opts) = @_;

    if ( ! defined $tests ) { return undef; }
    if ( ! defined $opts  ) { return undef; }

    $opts->{'REMOTE_FILES'}{'INETD_CONF'} = '/etc/xinetd.conf';
    $opts->{'LOCAL_FILES'}{'INETD_CONF'} = 
	obtain_remote_file($tests, $opts, 
			   $opts->{'REMOTE_FILES'}{'INETD_CONF'});

    {
	# Handle xinetd.conf's 'includedir' option
	# If found, get all the files in remote includedir
	# Ignore filenames containing '.' and ending in '~'
	#   as per xinetd.conf(5)
	my $fh = new IO::File;

	while ( 1 )
	{
	    if ( $fh->open("< $opts->{LOCAL_FILES}{INETD_CONF}") )
	    {
		my $includedir = undef;

		while ( my $line = $fh->getline )
		{
		    chomp($line);
		    if ( $line =~ m/\s*includedir\s+/ )
		    {
			$includedir = $line;
			$includedir =~ s/^\s+//;
			$includedir =~ s/^\S+\s+(\S+)\s*$/$1/;
			last;
		    }
		}

		$fh->close;
		if ( ! defined $includedir )
		{
		    # XXX FIXME?  Is this right?
		    $includedir = '/etc/xinetd.d';
		    last;
		}

		my @cmds;
		push @cmds, "for i in \\`ls -l /etc/xinetd.d | grep -v drwx | awk \'{print \\\$9}\' | grep -v -e \'~\$\' | grep -v -e \'\\\.\'\\`; do cat $includedir/\\\$i ; done";
		my $a = run_remote_cmds(\@cmds, $opts);

		if ( scalar(@$a) != 0 )
		{
		    if ( $a->[0]{'ExitStatus'} == 0 && 
			 $a->[0]{'RemoteExitStatus'} == 0 )
		    {
			my @replac = split "\n", $a->[0]{'Stdout'};

			my ($fhnew, $fnew) = 
			    tempfile(
				     TEMPLATE => 'remoteXXXXXX',
				     DIR      => $opts->{'TMPDIR'},
				     UNLINK   => 0,
				     );
			if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $fnew created\n"; }

			if ( $fh->open("< $opts->{LOCAL_FILES}{INETD_CONF}") )
			{
			    while ( my $line = $fh->getline )
			    {
				chomp($line);
				if ( $line =~ m/\s*includedir\s+$includedir/ )
				{
				    for ( my $i = 0 ; $i < scalar(@replac); 
					  $i++ )
				    {
					print $fhnew "$replac[$i]\n";
				    }
				}
				else
				{
				    print $fhnew "$line\n";
				}
			    }

			    $fh->close;
			    unlink $opts->{'LOCAL_FILES'}{'INETD_CONF'};
			    if ( defined $opts->{'DEBUG'} )
			    {
				print "DEBUG: Removing file $opts->{'LOCAL_FILES'}{'INETD_CONF'}\n";
			    }
			    $opts->{'LOCAL_FILES'}{'INETD_CONF'} = $fnew;
			}
		    }
		}
	    }
	}
    }

    if ( ! defined $opts->{'LOCAL_FILES'}{'INETD_CONF'} )
    {
	return undef;
    }

    return hashify_inet_conf($opts->{'REMOTE_FILES'}{'INETD_CONF'}, 
			     $opts->{'LOCAL_FILES'}{'INETD_CONF'});
}


#
# hashify_internet_services
#
# Return a hash reference representing internet services contained
# in the given /etc/services-like file.
#
sub hashify_internet_services($)
{
    my $f = shift;

    if ( ! defined $f ) { return undef; }
    if ( ! -f $f )      { return undef; }

    my %h;

    my $fh = new IO::File;

    if ( $fh->open("< $f") )
    {
	while ( my $line = $fh->getline )
	{
	    chomp($line);
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if    ( ! defined $line || length($line) == 0 ) { next;}
	    elsif ( $line =~ m/^\#/ ) { next; }

	    my $svc = $line;
	    $svc =~ s/^(\S+)\s+\d+\/\S+.*/$1/;
	    my $port = $line;
	    $port =~ s/^\S+\s+(\d+)\/\S+.*/$1/;
	    my $protocol = $line;
	    $protocol =~ s/^\S+\s+\d+\/(\S+).*/$1/;

	    $h{$svc}{$protocol}{'PORT'}    = $port;

	    my $aliases = $line;
	    $aliases =~ s/^\S+\s+\d+\/\S+\s*//;
	    if ( length($aliases) > 0 )
	    {
		$aliases =~ s/\#.*$//;
		$h{$svc}{$protocol}{'ALIASES'} = $aliases;
	    }
	}
    }

    return \%h;
}


#
# hashify_remote_internet_services
#
# Return a hash reference representing the remote site's /etc/services.
#
sub hashify_remote_internet_services($$)
{
    my ($tests, $opts) = @_;

    $opts->{'LOCAL_FILES'}{'SERVICES'} = 
	obtain_remote_file($tests, $opts, '/etc/services');

    return hashify_internet_services($opts->{'LOCAL_FILES'}{'SERVICES'});
}


#
# hashify_gatekeeper_conf
#
# Return a hash reference representing a globus gatekeeper conf file.
#
sub hashify_gatekeeper_conf($)
{
    my $f = shift;

    if ( ! defined $f ) { return undef; }
    if ( ! -f $f )      { return undef; }

    my %h;

    my $fh = new IO::File;

    if ( $fh->open("< $f") )
    {
	while ( my $line = $fh->getline )
	{
	    chomp($line);
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if    ( ! defined $line || length($line) == 0 ) { next;}
	    elsif ( $line =~ m/^\#/ ) { next; }

	    my $key = $line;
	    $key =~ s/^(\S+)\s+.*$/$1/;
	    my $val = undef;
	    if ( $line =~ /^(\S+)\s+\S+.*$/ )
	    {
		$val = $line;
		$val =~ s/^\S+\s+(.*)$/$1/;
	    }

	    $h{$key} = $val;
	}
    }

    return \%h;
}


#
# hashify_jobmanager
#
# Return a hash reference representing a jobmanager.
#
sub hashify_jobmanager($)
{
    my $f = shift;

    if ( ! defined $f ) { return undef; }
    if ( ! -f $f )      { return undef; }

    my %h;

    my $fh = new IO::File;

    if ( $fh->open("< $f") )
    {
	while ( my $line = $fh->getline )
	{
	    chomp($line);
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if    ( ! defined $line || length($line) == 0 ) { next;}
	    elsif ( $line =~ m/^\#/ ) { next; }

	    # Everything after the 4th argument is an argument
	    # supplied to the globus-job-manager - this is what we care about
	    my @tokens = split /\s+/, $line;

	    my $num_args = 4;
	    if ( scalar(@tokens) < $num_args ) { $num_args = scalar(@tokens); }

	    splice @tokens, 0, $num_args;

	    # globus-job-manager options begin with a dash
	    # and the options may have args.

	    # XXX - FIXME - the following is easily fooled, but should 
	    # XXX - FIXME - get the job done at the level we care.
	    my $opt = undef;
	    foreach my $token (@tokens)
	    {
		if ( $token =~ /^\-/ )
		{
		    if ( defined $opt ) { $h{$opt} = undef; }
		    $opt = $token;
		}
		else
		{
		    if ( defined $opt ) { $h{$opt} = $token; }
		    $opt = undef;
		}
	    }

	    if ( defined $opt ) { $h{$opt} = undef; }
	}
    }

    return \%h;
}


#
# hashify_ml_properties
#
# Return a hash reference representing a MonALISA ml.properties file.
#
sub hashify_ml_properties($)
{
    my $f = shift;

    if ( ! defined $f ) { return undef; }
    if ( ! -f $f )      { return undef; }

    my %h;

    my $fh = new IO::File;

    if ( $fh->open("< $f") )
    {
	while ( my $line = $fh->getline )
	{
	    chomp($line);
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if    ( ! defined $line || length($line) == 0 ) { next;}
	    elsif ( $line =~ m/^\#/ ) { next; }

	    my @tokens = split "=", $line;
	    my $key = shift @tokens;
	    my $val = $tokens[0];
	    if ( scalar(@tokens) > 1 ) { $val = join "=", @tokens; }

	    $key =~ s/^\s+//;
	    $key =~ s/\s+$//;

	    $val =~ s/^\s+//;
	    $val =~ s/\s+$//;

	    $h{$key} = $val;
	}
    }

    return \%h;
}


#
# get_grid3_info_var
#
# Return the value of the requested variable from grid3-info.conf.
#
sub get_grid3_info_var($$$)
{
    my ($tests, $opts, $var) = @_;

    if ( ! defined $tests ) { return undef; }
    if ( ! defined $opts )  { return undef; }
    if ( ! defined $var )   { return undef; }

    my $cmd = "/bin/sh -c '. $opts->{LOCAL_FILES}{OSG_ATTRIBUTES_CONF} > /dev/null 2>&1 ; echo \$$var'";
    my $d = `$cmd`;

    if ( defined $d && length($d) > 0 ) { chomp($d); }
    if ( defined $d && length($d) == 0 ) { $d = undef; }

    return $d;
}


#
# get_grid3_setup_var
#
# Return the value of the requested variable from $GRID3/setup.sh.
#
sub get_grid3_setup_var($$$)
{
    my ($tests, $opts, $var) = @_;

    if ( ! defined $tests ) { return undef; }
    if ( ! defined $opts )  { return undef; }
    if ( ! defined $var )   { return undef; }

    my $cmd = "/bin/sh -c '. $opts->{LOCAL_FILES}{GRID3_SETUP_SH} > /dev/null 2>&1 ; echo \$$var'";
    my $d = `$cmd`;

    if ( defined $d && length($d) > 0 ) { chomp($d); }
    if ( defined $d && length($d) == 0 ) { $d = undef; }

    return $d;
}


#
# get_grid_info_var
#
# Return the value of the requested variable from grid-info.conf.
#
sub get_grid_info_var($$$)
{
    my ($tests, $opts, $var) = @_;

    if ( ! defined $tests ) { return undef; }
    if ( ! defined $opts )  { return undef; }
    if ( ! defined $var )   { return undef; }

    my $cmd = "/bin/sh -c '. $opts->{LOCAL_FILES}{GRID_INFO_CONF} > /dev/null 2>&1 ; echo \$$var'";
    my $d = `$cmd`;

    if ( defined $d && length($d) > 0 ) { chomp($d); }
    if ( defined $d && length($d) == 0 ) { $d = undef; }

    return $d;
}


#
# get_grid_mapfile_from_gatekeeper_conf 
#
# Return the grid-mapfile referred to in the globus-gatekeeper.conf
#
sub get_grid_mapfile_from_gatekeeper_conf($)
{
    my $ctx = shift;

    my $grid_mapfile = undef;

    if ( exists $ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'} )
    {
	if ( exists $ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-gridmap'} &&
	     defined $ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-gridmap'} )
	{
	    $grid_mapfile = 
		$ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-gridmap'};
	}
    }

    return $grid_mapfile;
}


#
# get_grid_services_from_gatekeeper_conf 
#
# Return the grid-services location referred to in the globus-gatekeeper.conf
#
sub get_grid_services_from_gatekeeper_conf()
{
    my ($tests, $ctx)  = @_;

    # Find where the grid_services live
    my $grid_services = undef;

    if ( exists $ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'} )
    {
	if ( exists $ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-grid_services'} &&
	     defined $ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-grid_services'} )
	{
	    my $tmp = 
		$ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-grid_services'};

	    if ( $tmp !~ /^\// )
	    {
		if ( exists $ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-home'} &&
		     defined $ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-home'} )
		{
		    $grid_services = 
			$ctx->{'FILE_HASH'}{'GATEKEEPER_CONF'}->{'-home'};
		}
		else
		{
		    # use existing GLOBUS_LOCATION
		    $grid_services = $tests->{'RESULTS'}{'globus_location'};
		}

		if ( $grid_services !~ /\/$/ ) { $grid_services .= '/'; }

		$grid_services .= $tmp;	    
	    }
	    else
	    {
		$grid_services = $tmp;
	    }
	}
    }

    return $grid_services;
}


#
# check_prereqs
#
# Return 0 if prereq tests all passed, non-zero if not.
#
sub check_prereqs($$)
{
    my ($tests, $opts)  = @_;
    my $t = undef;
    if ( exists $opts->{'TESTNAME'} )
    {
	$t = $opts->{'TESTNAME'}
    }

    if ( defined $tests->{'PREREQS'}{$t} )
    {
	my @prereqs;

	if ( $tests->{'PREREQS'}{$t} =~ m/,/ )
	{
	    @prereqs = split ',', $tests->{'PREREQS'}{$t};
	}
	else
	{
	    push @prereqs, $tests->{'PREREQS'}{$t};
	}

	for ( my $i = 0; $i < scalar(@prereqs); $i++ )
	{
	    if ( ! defined $tests->{'STATUS'}{$prereqs[$i]} ||
		 $tests->{'STATUS'}{$prereqs[$i]} eq FAILURE ||
		 $tests->{'STATUS'}{$prereqs[$i]} eq UNTESTED )
	    {
		return 1;
	    }
	}
    }

    return 0;
}


#
# pre_check
#
# Generic pre-test routine.  Just print out "Checking...".
#
sub pre_check($$)
{
    my ($tests, $opts)  = @_;
    my $t = undef;
    if ( exists $opts->{'TESTNAME'} )
    {
	$t = $opts->{'TESTNAME'}
    }

    if ( defined $tests->{'DESCRIP'}{$t} )
    {
	print 'Checking ' . $tests->{'DESCRIP'}{$t} . ': ';
    }
    else
    {
	print "Running test \"$t\": ";
    }
}


#
# post_check
#
# Generic post-test routine.  Just print out the result.
#
sub post_check($$)
{
    my ($tests, $opts)  = @_;
    my $t = undef;
    if ( exists $opts->{'TESTNAME'} )
    {
	$t = $opts->{'TESTNAME'}
    }

    if ( defined $tests->{'RESULTS'}{$t} )
    {
	$tests->{'RESULTS'}{$t} =~ s/\n$//;
	print $tests->{'RESULTS'}{$t} . "\n";
    }
    elsif ( defined $tests->{'STATUS'}{$t} )
    {
	$tests->{'STATUS'}{$t} =~ s/\n$//;
	print $tests->{'STATUS'}{$t} . "\n";

	if ( $tests->{'STATUS'}{$t} eq SUCCESS )
	{
	    if ( defined $tests->{'ACTONPASS'}{$t} )
	    {
		&$tests->{'ACTONPASS'}{$t}($tests, $opts);
	    }
	}
	elsif ( $tests->{'STATUS'}{$t} eq FAILURE )
	{
	    if ( defined $tests->{'ACTONFAIL'}{$t} )
	    {
		&{$tests->{'ACTONFAIL'}{$t}}($tests, $opts);
	    }
	}
    }
    else
    {
	print UNKNOWN . "\n";
    }
}


#
# print_msgs_and_exit
#
# Print any messages corresponding to the test and exit.
#
sub print_msgs_and_exit($$)
{
    my ($tests, $opts)  = @_;

    if ( defined $tests->{'MESGS'}{$opts->{'TESTNAME'}} )
    {
	print $tests->{'MESGS'}{$opts->{'TESTNAME'}} . "\n";
    }

    exit;
}


#
# check_path
#
# Make sure all needed programs are in the user's PATH.
#
sub check_path($$)
{
    my ($tests, $opts)  = @_;
    my $t = undef;
    if ( exists $opts->{'TESTNAME'} )
    {
	$t = $opts->{'TESTNAME'}
    }

    $tests->{'MESGS'}{$t}   = undef;
    $tests->{'STATUS'}{$t} = UNTESTED;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my @exes = (
		'grid-proxy-info',
		'globusrun',
		'globus-job-run',
		'globus-url-copy',
		);

    my @not_found;

    foreach my $exe (@exes)
    {
	if ( system("/bin/sh -c \"type $exe > /dev/null 2>&1\"") != 0 )
	{
	    push @not_found, $exe;
	}
    }

    if ( scalar(@not_found) == 0 )
    {
	$tests->{'MESGS'}{$t}   = 'All prerequisites needed for testing were found.';
	$tests->{'STATUS'}{$t} = SUCCESS;
    }
    else
    {
	$tests->{'MESGS'}{$t} = 'Error: Can\'t find prerequisites: ' . 
	    join ',',@not_found;
	$tests->{'STATUS'}{$t} = FAILURE;
    }
}

#
# check_proxy
#
# Check for a valid user proxy.
#
sub check_proxy($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my @cmds;
    push @cmds, "grid-proxy-info -exists";

    my $a = run_cmds(\@cmds, $opts);

    if ( scalar(@$a) != 0 )
    {
	if ( $a->[0]{'ExitStatus'} == 0 )
	{
	    $tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
	}
    }
    else
    {
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = 
	    'Error: No proxy exists for ' . 
	    $opts->{'USER'} . '@' . $opts->{'LOCALHOST'};
    }
}


#
# check_remote_host_reachable
#
# Check if we can see the remote host.
#
sub check_remote_host_reachable
{
    my ($tests, $opts)  = @_;

    my $host = undef;
    if ( exists $opts->{'GATEKEEPER_HOST'} )
    {
	$host = $opts->{'GATEKEEPER_HOST'};
    }

    my $port = undef;
    if ( exists $opts->{'GATEKEEPER_PORT'} )
    {
	$port = $opts->{'GATEKEEPER_PORT'};
    }

    my $timeout = 10;
    if ( exists $opts->{'TIMEOUT'} )
    {
	$timeout = $opts->{'TIMEOUT'};
    }

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    if ( ! defined $host )
    {
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = 'Host not supplied';
	return;
    }

    if ( ! defined $port )
    {
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = 'Port not supplied';
	return;
    }


    # Try a few ports
    my @ports = ($port, 22, 23);

    # Try to connect via tcp to $host:$port
    for ( my $i = 0; $i < scalar(@ports); $i++ )
    {
	my $sock = IO::Socket::INET->new(Proto    => "tcp",
					 PeerAddr => "$host:$ports[$i]",
					 Timeout  => $timeout,
					 );
	if ( defined $sock )
	{
	    $sock->autoflush(1);
	    close($sock);
	    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
	    return;
	}
    }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    $tests->{'MESGS'}{$opts->{'TESTNAME'}} = NO . "\n  Can't connect to $host";
}


#
# check_gatekeeper_service
#
# Check to see if remote gatekeeper is running.
#
sub check_gatekeeper_service($$)
{
    my ($tests, $opts)  = @_;

    my $host = undef;
    if ( exists $opts->{'GATEKEEPER_HOST'} )
    {
	$host = $opts->{'GATEKEEPER_HOST'};
    }

    my $port = 2119;
    if ( exists $opts->{'GATEKEEPER_PORT'} )
    {
	$port = $opts->{'GATEKEEPER_PORT'};
    }

    my $timeout = 10;
    if ( exists $opts->{'TIMEOUT'} )
    {
	$timeout = $opts->{'TIMEOUT'};
    }

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    if ( ! defined $host )
    {
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = 'Host not supplied';
	return;
    }

    my $sock = new IO::Socket::INET (Timeout  => $timeout,
				     PeerAddr => "$host:$port",
				     Proto    => "tcp",
				     );

    if ( ! defined $sock || ! defined $sock->connected )
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = NO . 
	    "\n  Can't connect to $host:$port/tcp";
	return;
    }

    $sock->autoflush(1);
    close($sock);

    $tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = YES . "; port $port";
}


#
# check_auth
#
# Check if the user can authenticate at remote site.
#
sub check_auth($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $host = undef;
    if ( exists $opts->{'GATEKEEPER_HOST'} )
    {
	$host = $opts->{'GATEKEEPER_HOST'};
    }

    my $port = 2119;
    if ( exists $opts->{'GATEKEEPER_PORT'} )
    {
	$port = $opts->{'GATEKEEPER_PORT'};
    }

    if ( ! defined $host )
    {
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = 'Host not supplied';
	return;
    }

    my $dest = $host;
    if ( defined $port ) { $dest .= ":$port" }

    my @cmds;
    push @cmds, "globusrun -a -r $dest";

    my $a = run_cmds(\@cmds, $opts);

    if ( scalar(@$a) != 0 )
    {
	if ( $a->[0]{'ExitStatus'} == 0 )
	{
	    $tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
	}
    }
}


#
# check_hello_world
#
# Check to see if we can run a trivial job via the fork manager.
#
sub check_hello_world($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $str = 'Hello, World';

    my @testnames;
    my @cmds;
    push @cmds, "echo \'$str\'";
    push @testnames, $opts->{'TESTNAME'};

    if ( defined $opts->{'CACHING_ENABLED'} )
    {
	# Optimization - run some other tests in advance
	push @cmds, 'uptime';
	push @testnames, 'uptime';
	push @cmds, 'cat /etc/services';
	push @testnames, 'services';
	push @cmds, 'echo \$GLOBUS_LOCATION';
	push @testnames, 'globus_location';
    }

    my $a = run_remote_cmds(\@cmds, $opts);

    for ( my $i = 0; $i < scalar(@$a); $i++ )
    {
	if ( $a->[$i]{'ExitStatus'} == 0 && $a->[$i]{'RemoteExitStatus'} == 0 )
	{
	    if ( $testnames[$i] eq 'uptime' )
	    {
		$tests->{'RESULTS'}{$testnames[$i]} = SUCCESS . "\n  " . 
			$a->[$i]{'Stdout'};
	    }
	    elsif ( $testnames[$i] eq 'services' )
	    {
		$opts->{'LOCAL_FILES'}{'SERVICES'} = 
		    stdout_to_file($a->[$i]{'Stdout'}, $opts);
	    }
	    elsif ( $testnames[$i] eq 'globus_location' )
	    {
		$tests->{'RESULTS'}{$testnames[$i]} = $a->[$i]{'Stdout'};
	    }
	}
    }
}


#
# check_uptime
#
# Check uptime of remote gatekeeper.
#
sub check_uptime($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

	my @cmds;
	push @cmds, "uptime";

	my $a = run_remote_cmds(\@cmds, $opts);

	if ( scalar(@$a) != 0 )
	{
	    if ( $a->[0]{'ExitStatus'} == 0 && $a->[0]{'RemoteExitStatus'} == 0 )
	    {
		$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = SUCCESS . "\n  " . 
		    $a->[0]{'Stdout'};
		$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
		$tests->{'STDOUT'}{$opts->{'TESTNAME'}} = $a->[0]{'Stdout'};
		$tests->{'STDERR'}{$opts->{'TESTNAME'}} = $a->[0]{'Stderr'};
	    }
	}
    }
}


#
# check_hostcert
#
# Obtain remote hostcert.pem and check validity.
#
sub check_hostcert($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

	my $globus_location = $tests->{'RESULTS'}{'globus_location'};
	my @cmds;
	push @cmds, "openssl=\\\"openssl\\\"; if test -x $globus_location/bin/openssl; then openssl=\\\"$globus_location/bin/openssl\\\"; if test -z \\\$LD_LIBRARY_PATH; then LD_LIBRARY_PATH=\\\"$globus_location/lib\\\"; else LD_LIBRARY_PATH=\\\"$globus_location/lib:\\\$LD_LIBRARY_PATH\\\"; fi; export LD_LIBRARY_PATH; fi; \\\$openssl x509 -in /etc/grid-security/hostcert.pem -noout -text | grep -e \'Not\\ \\+After\'";

	my $a = run_remote_cmds(\@cmds, $opts);

	if ( scalar(@$a) != 0 )
	{
	    if ( $a->[0]{'ExitStatus'} == 0 && $a->[0]{'RemoteExitStatus'} == 0 )
	    {
		(my $r = $a->[0]{'Stdout'}) =~ s/^\s*Not\s+After\s+\:\s+//;

		$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $r;
		$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
		$tests->{'STDOUT'}{$opts->{'TESTNAME'}} = $a->[0]{'Stdout'};
		$tests->{'STDERR'}{$opts->{'TESTNAME'}} = $a->[0]{'Stderr'};
	    }
	}
    }
}


#
# check_services
#
# Get a copy of the remote /etc/services.
#
sub check_services($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    if ( ! exists $opts->{'FILE_HASH'}{'SERVICES'} )
    {
	if ( ! exists $opts->{'LOCAL_FILES'}{'SERVICES'} )
	{
	    $opts->{'FILE_HASH'}{'SERVICES'} = 
		hashify_remote_internet_services($tests, $opts);
	}
	else
	{
	    $opts->{'FILE_HASH'}{'SERVICES'} = 
		hashify_internet_services($opts->{'LOCAL_FILES'}{'SERVICES'});
	}
    }

    if ( ! exists  $opts->{'FILE_HASH'}{'SERVICES'} || 
	 ! defined $opts->{'FILE_HASH'}{'SERVICES'} ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
}


#
# check_inetd
#
# Analyze the config of the remote INET super-server.
#
sub check_inetd($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    # Find the service config for this service name

    if ( ! exists $opts->{'FILE_HASH'}{'INETD_CONF'} )
    {
	if ( ! exists $opts->{'LOCAL_FILES'}{'INETD_CONF'} )
	{
	    $opts->{'FILE_HASH'}{'INETD_CONF'} = 
		hashify_remote_inet_conf($tests, $opts);
	}
	else
	{
	    $opts->{'FILE_HASH'}{'INETD_CONF'} = 
		hashify_inet_conf($opts->{'REMOTE_FILES'}{'INETD_CONF'}, 
				  $opts->{'LOCAL_FILES'}{'INETD_CONF'});
	}
    }

    if ( ! exists  $opts->{'FILE_HASH'}{'INETD_CONF'} ||
	 ! defined $opts->{'FILE_HASH'}{'INETD_CONF'} ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
}


#
# check_gsiftp_service
#
# Check to see if remote gsiftp server is running.
#
sub check_gsiftp_service($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $host = undef;
    if ( exists $opts->{'GSIFTP_HOST'} )
    {
	$host = $opts->{'GSIFTP_HOST'};
    }
    elsif ( exists $opts->{'GATEKEEPER_HOST'} )
    {
	$host = $opts->{'GATEKEEPER_HOST'};
    }

    if ( ! defined $host )
    {
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = 'Host not supplied';
	return;
    }

    # If gsiftp port is supplied, use it.
    #
    # Otherwise, 
    #
    # Figure out the service name corresponding to gridftp from the inetd.conf
    # Assume the 'correct' gsiftp has the same GLOBUS_LOCATION
    my $port = undef;
    if ( exists $opts->{'GSIFTP_PORT'} )
    {
	$port = $opts->{'GSIFTP_PORT'};
    }
    else
    {
	# XXX  Beware.  Heuristics ahead.
	#
	# It can happen that the correct gsiftp server port cannot
	# be deduced by what follows - that is, if there is more than
	# one valid path to the grid-mapfile or GLOBUS_LOCATION
	# (via soft linking or alternate mountpoints).  This
	# should be improved.
	my $globus_location = $tests->{'RESULTS'}{'globus_location'};

	my $grid_mapfile = get_grid_mapfile_from_gatekeeper_conf($opts);

	my $svcname = undef;

	if ( exists  $opts->{'FILE_HASH'}{'INETD_CONF'} &&
	     defined $opts->{'FILE_HASH'}{'INETD_CONF'} )
	{
	    my $href = $opts->{'FILE_HASH'}{'INETD_CONF'};

	    # Assume tcp
	    foreach my $svc ( keys %$href )
	    {
		if ( ! exists $href->{$svc}{'tcp'}{'server'} ) { next; }

		if ( $href->{$svc}{'tcp'}{'server'} =~ m/in\.ftpd/ )
		{
		    if ( exists $href->{$svc}{'tcp'}{'env'} &&
			 ($href->{$svc}{'tcp'}{'env'} =~ 
			  m/GRIDMAP=$grid_mapfile/) )
		    {
			$svcname = $svc;
		    }
		    elsif ( exists $href->{$svc}{'tcp'}{'env'} &&
			 ($href->{$svc}{'tcp'}{'env'} =~ 
			  m/LD_LIBRARY_PATH=$globus_location/) )
		    {
			$svcname = $svc;
		    }
		    elsif ( exists $href->{$svc}{'tcp'}{'server_args'} &&
			    ($href->{$svc}{'tcp'}{'server_args'} =~ 
			     m/-G\s+$globus_location/) )
		    {
			$svcname = $svc;
		    }
		    elsif ( exists $href->{$svc}{'tcp'}{'server'} )
		    {
			my $name1 = $href->{$svc}{'tcp'}{'server'};
			my $name2 = "$globus_location/sbin/in.ftpd";

			$name1 =~ s/\/+/\//g;
			$name2 =~ s/\/+/\//g;

			if ( $name1 eq $name2 ) { $svcname = $svc; }
		    }

		    if ( defined $svcname ) { last; }
		}
	    }
	}

	# If we've found the service name, 
	# map it to the port given in /etc/services
	if ( defined $svcname )
	{
	    $opts->{'SVCNAMES'}{'gsiftp'} = $svcname;

	    if ( ! exists $opts->{'FILE_HASH'}{'SERVICES'} )
	    {
		if ( ! exists $opts->{'LOCAL_FILES'}{'SERVICES'} )
		{
		    $opts->{'FILE_HASH'}{'SERVICES'} = 
			hashify_remote_internet_services($tests, $opts);
		}
		else
		{
		    $opts->{'FILE_HASH'}{'SERVICES'} = 
			hashify_internet_services($opts->{'LOCAL_FILES'}{'SERVICES'});
		}
	    }

	    # XXX FIXME XXX protocol could be an /etc/protocols alias
	    if ( exists $opts->{'FILE_HASH'}{'SERVICES'}{$svcname}{'tcp'}{'PORT'} )
	    {
		$port = $opts->{'FILE_HASH'}{'SERVICES'}{$svcname}{'tcp'}{'PORT'};
	    }
	}
    }

    # Last resort - guess...
    if ( ! defined $port ) { $port = 2811; }

    $opts->{'GSIFTP_PORT'} = $port;

    my $timeout = 10;
    if ( exists $opts->{'TIMEOUT'} )
    {
	$timeout = $opts->{'TIMEOUT'};
    }

    my $sock = new IO::Socket::INET (Timeout  => $timeout,
				     PeerAddr => "$host:$port",
				     Proto    => "tcp",
				     );

    if ( ! defined $sock || ! defined $sock->connected )
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = NO . 
	    "\n  Can't connect to $host:$port/tcp";
	return;
    }

    $sock->autoflush(1);
    close($sock);

    $tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = YES . "; port $port";
}


# TODO: Run GIP validator as part of site-verify


#
# check_gsiftp_local_out
#
# Check to see if we can use the local gridftp client to push a file to
# the remote site.
#
sub check_gsiftp_local_out($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $host = undef;
    if ( exists $opts->{'GSIFTP_HOST'} )
    {
	$host = $opts->{'GSIFTP_HOST'};
    }
    elsif ( exists $opts->{'GATEKEEPER_HOST'} )
    {
	$host = $opts->{'GATEKEEPER_HOST'};
    }

    my $port = 2811;
    if ( exists $opts->{'GSIFTP_PORT'} )
    {
	$port = $opts->{'GSIFTP_PORT'};
    }

    if ( ! defined $host )
    {
	$tests->{'MESGS'}{$opts->{'TESTNAME'}} = 'Host not supplied';
	return;
    }

    my $dest = $host;
    if ( defined $port ) { $dest .= ":$port" }

    my ($f, $fname) = tempfile(
			       TEMPLATE => 'gsiftpXXXXXX',
			       DIR      => $opts->{'TMPDIR'}, 
			       SUFFIX   => '.local_out', 
			       UNLINK   => 0,
			       );
    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $fname created\n"; }

    print $f "local: " . $opts->{'LOCALHOST'} . " <-> $host\n";

    my ($base, $path) = fileparse($fname);

    $opts->{'LOCAL_GSIFTP_OUTFILE'} = $fname;

    my $rfname = "/tmp/${base}_moved";

    $opts->{'REMOTE_GSIFTP_OUTFILE'} = $rfname;

    my @cmds;
    push @cmds, "globus-url-copy file:$fname gsiftp://$dest/$rfname";

    my $a = run_cmds(\@cmds, $opts);

    if ( $a->[0]{'ExitStatus'} == 0 )
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}


#
# check_gsiftp_local_in
#
# Check to see if we can use the local gridftp client to pull a file from
# the remote site.
#
sub check_gsiftp_local_in($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    $opts->{'GSIFTP_ONLY'} = 1;

    $opts->{'LOCAL_GSIFTP_INFILE'} = 
	obtain_remote_file($tests, $opts, $opts->{'REMOTE_GSIFTP_OUTFILE'});

    delete $opts->{'GSIFTP_ONLY'};

    if ( defined $opts->{'LOCAL_GSIFTP_INFILE'} )
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}


#
# check_gsiftp_local_diffs
#
# Verify there are no differences in files transferred with local gsiftp client
#
sub check_gsiftp_local_diffs($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $f1 = undef;
    if ( exists $opts->{'LOCAL_GSIFTP_OUTFILE'} )
    {
	$f1 = $opts->{'LOCAL_GSIFTP_OUTFILE'};
    }

    my $f2 = undef;
    if ( exists $opts->{'LOCAL_GSIFTP_INFILE'} )
    {
	$f2 = $opts->{'LOCAL_GSIFTP_INFILE'};
    }

    if ( ! defined $f1 ) { return; }
    if ( ! defined $f2 ) { return; }
    if ( ! -e $f1 ) { return; }
    if ( ! -e $f2 ) { return; }

    if ( compare("$f1", "$f2") == 0 )
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;

	# cleanup
	unlink $opts->{'LOCAL_GSIFTP_OUTFILE'};
	if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $opts->{'LOCAL_GSIFTP_OUTFILE'}\n"; }
	unlink $opts->{'LOCAL_GSIFTP_INFILE'};
	if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $opts->{'LOCAL_GSIFTP_INFILE'}\n"; }

	my @cmds;
	push @cmds, "rm -f $opts->{REMOTE_GSIFTP_OUTFILE}";

	run_remote_cmds(\@cmds, $opts);

	delete $opts->{'LOCAL_GSIFTP_INFILE'};
	delete $opts->{'LOCAL_GSIFTP_OUTFILE'};
	delete $opts->{'REMOTE_GSIFTP_OUTFILE'};
    }
}


#
# check_globus_location
#
# Determine the GLOBUS_LOCATION definition of the remote host.
# We do it by tracking the gatekeeper service.
#
sub check_globus_location($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

	# We first try to the simple way: echo GLOBUS_LOCATION in a fork job.
	# If that doesn't work, analyze the remote site's /etc/{services,xinetd}
	my @cmds;
	push @cmds, 'echo \$GLOBUS_LOCATION';
	my $a = run_remote_cmds(\@cmds, $opts);

	if ( scalar(@$a) != 0 )
	{
	    if ( $a->[0]{'ExitStatus'} == 0 && $a->[0]{'RemoteExitStatus'} == 0 )
	    {
		$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $a->[0]{'Stdout'};
		$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
		$tests->{'STDOUT'}{$opts->{'TESTNAME'}} = $a->[0]{'Stdout'};
		$tests->{'STDERR'}{$opts->{'TESTNAME'}} = $a->[0]{'Stderr'};
		return;
	    }
	}

	if ( ! exists  $opts->{'FILE_HASH'}{'SERVICES'} ||
	     ! defined $opts->{'FILE_HASH'}{'SERVICES'} ) { return; }

	my $port = 2119;
	if ( exists $opts->{'GATEKEEPER_PORT'} )
	{
	    $port = $opts->{'GATEKEEPER_PORT'};
	}

	if ( ! defined $port ) { return; }

	# get the service name for the port we are using
	my $svcname = get_service_name($opts, $port, 'tcp');
	if ( ! defined $svcname ) { return; }

	# Find the service config for this service name
	my $href = get_inetd_service_name($opts, $svcname, 'tcp');

	if ( ! defined $href ) { return; }

	my $globus_location = undef;

	if ( exists $href->{'env'} && ($href->{'env'} =~ m/LD_LIBRARY_PATH/) )
	{
	    $globus_location = $href->{'env'};
	    $globus_location =~ s/^.*LD_LIBRARY_PATH=//;
	    $globus_location =~ s/(\S+)\s+.*$/$1/;
	    my ($base, $path) = fileparse($globus_location);
	    $globus_location = $path;
	}
	elsif ( exists $href->{'server'} )
	{
	    my ($base,  $path)  = fileparse($href->{'server'});
	    my ($base1, $path1) = fileparse($path);
	    $globus_location = $path1;
	}

	if ( ! defined $globus_location ) { return; }

	$globus_location =~ s/\/$//;

	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $globus_location;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }

    if ( defined $opts->{'CACHING_ENABLED'} )
    {
	if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq SUCCESS )
	{
	    my $globus_location = $tests->{'RESULTS'}{$opts->{'TESTNAME'}};

	    my $rfile = $globus_location . 
		'/libexec/globus-sh-tools-vars.sh';

	    my @testnames;
	    my @cmds;
	    push @cmds, "cat $rfile";
	    push @testnames, 'globus_sh_tools_vars';
	    push @cmds, "openssl=\\\"openssl\\\"; if test -x $globus_location/bin/openssl; then openssl=\\\"$globus_location/bin/openssl\\\"; if test -z \\\$LD_LIBRARY_PATH; then LD_LIBRARY_PATH=\\\"$globus_location/lib\\\"; else LD_LIBRARY_PATH=\\\"$globus_location/lib:\\\$LD_LIBRARY_PATH\\\"; fi; export LD_LIBRARY_PATH; fi; \\\$openssl x509 -in /etc/grid-security/hostcert.pem -noout -text | grep -e \'Not\\ \\+After\'";
	    push @testnames, 'hostcert';
	    push @cmds, "cat $globus_location/etc/grid-info.conf";
	    push @testnames, 'grid_info';
	    push @cmds, "if test -x $globus_location/../vdt/bin/vdt-version ; then VDT_LOCATION=$globus_location/.. $globus_location/../vdt/bin/vdt-version; else rpm -qf $globus_location/bin/openssl; fi";
	    push @testnames, 'vdt_deployment';

	    my $a = run_remote_cmds(\@cmds, $opts);

	    for ( my $i = 0; $i < scalar(@$a); $i++ )
	    {
		if ( $a->[$i]{'ExitStatus'} == 0 && 
		     $a->[$i]{'RemoteExitStatus'} == 0 )
		{
		    if ( $testnames[$i] eq 'hostcert' )
		    {
			(my $r = $a->[$i]{'Stdout'}) =~ s/^\s*Not\s+After\s+\:\s+//;
			$tests->{'STATUS'}{$testnames[$i]}  = SUCCESS;
			$tests->{'STDOUT'}{$testnames[$i]} = $a->[$i]{'Stdout'};
			$tests->{'STDERR'}{$testnames[$i]} = $a->[$i]{'Stderr'};
			$tests->{'RESULTS'}{$testnames[$i]} = $r;
		    }
		    elsif ( $testnames[$i] eq 'globus_sh_tools_vars' )
		    {
			$opts->{'LOCAL_FILES'}{'GLOBUS_SH_TOOLS_VARS'} = 
			    stdout_to_file($a->[$i]{'Stdout'}, $opts);
		    }
		    elsif ( $testnames[$i] eq 'grid_info' )
		    {
			$opts->{'LOCAL_FILES'}{'GRID_INFO_CONF'} = 
			    stdout_to_file($a->[$i]{'Stdout'}, $opts);
		    }
		    elsif ( $testnames[$i] eq 'vdt_deployment' )
		    {
			$opts->{'LOCAL_FILES'}{'VDT_VERSION_CHECK'} = 
			    stdout_to_file($a->[$i]{'Stdout'}, $opts);
		    }
		}
	    }
	}
    }
}


#
# check_gatekeeper_conf
#
# Find and retrieve the remote gatekeeper configuration file.
#
sub check_gatekeeper_conf($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $svcname = undef;

    if ( exists $opts->{'SVCNAMES'}{'globus-gatekeeper'} )
    {
	$svcname = $opts->{'SVCNAMES'}{'globus-gatekeeper'};
    }
    else
    {
	# See if we can figure out the service name
	if ( exists $opts->{'FILE_HASH'}{'INETD_CONF'} )
	{
	    my $hi = $opts->{'FILE_HASH'}{'INETD_CONF'};

	    if ( exists $opts->{'FILE_HASH'}{'SERVICES'} )
	    {
		my $hs = $opts->{'FILE_HASH'}{'SERVICES'};

		foreach my $svc ( keys %$hi )
		{
		    if ( exists $hs->{$svc}{'tcp'}{'PORT'} &&
			 ($hs->{$svc}{'tcp'}{'PORT'} == 
			  $opts->{'GATEKEEPER_PORT'}) )
		    {
			$opts->{'SVCNAMES'}{'globus-gatekeeper'} = $svc;
			$svcname = $svc;
			last;
		    }
		}
	    }

	    if ( ! defined $svcname )
	    {
		my $foo = $tests->{'RESULTS'}{'globus_location'};
		$foo =~ s/\//\\\/\+/g;

		# See if we can match up an inetd service via GLOBUS_LOCATION
		foreach my $svc ( keys %$hi )
		{
		    if ( exists $hi->{$svc}{'tcp'}{'server_args'} &&
			 $hi->{$svc}{'tcp'}{'server_args'} =~ m/^\-conf\s+$foo\/+etc\// )
		    {
			$svcname = $svc;
			$opts->{'SVCNAMES'}{'globus-gatekeeper'} = $svc;
			last;
			
		    }
		    elsif ( exists $hi->{$svc}{'tcp'}{'server'} &&
			    $hi->{$svc}{'tcp'}{'server'} =~ m/^$foo\/+sbin\/+globus\-gatekeeper$/ )
		    {
			$svcname = $svc;
			$opts->{'SVCNAMES'}{'globus-gatekeeper'} = $svc;
			last;
		    }
		}
	    }
	}
    }

    my $globus_gatekeeper_conf = undef;

    if ( defined $svcname )
    {
	if ( exists $opts->{'FILE_HASH'}{'INETD_CONF'}{$svcname}{'tcp'}{'server_args'} )
	{
	    $globus_gatekeeper_conf =
		$opts->{'FILE_HASH'}{'INETD_CONF'}{$svcname}{'tcp'}{'server_args'};
	    $globus_gatekeeper_conf =~ s/^.*\-conf\s+(\S+).*$/$1/;
	}
    }

    if ( ! defined $globus_gatekeeper_conf )
    {
	# Educated guess...
	$globus_gatekeeper_conf = $tests->{'RESULTS'}{'globus_location'} . 
	    '/etc/globus-gatekeeper.conf';
    }

    my $f = obtain_remote_file($tests, $opts, $globus_gatekeeper_conf);

    if ( ! defined $f || ! -f $f ) { print "Unable to retrieve $globus_gatekeeper_conf.\n"; return; }

    $opts->{'FILE_HASH'}{'GATEKEEPER_CONF'} = hashify_gatekeeper_conf($f);

    unlink $f;
    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $f\n"; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = YES . 
	"\n  $globus_gatekeeper_conf";

    if ( defined $opts->{'CACHING_ENABLED'} )
    {
	# Find where the grid-mapfile lives
	my $grid_mapfile = get_grid_mapfile_from_gatekeeper_conf($opts);

	# If we can't find it, guess
	if ( ! defined $grid_mapfile )
	{
	    $grid_mapfile = '/etc/grid-security/grid-mapfile';
	}

	# Find where the grid-services are located
	my $grid_services = get_grid_services_from_gatekeeper_conf();

	# If we can't find it, guess
	if ( ! defined $grid_services )
	{
	    $grid_services = $tests->{'RESULTS'}{'globus_location'};
	    if ( $grid_services !~ /\/$/ ) { $grid_services .= '/'; }	    
	    $grid_services .= 'etc/grid-services';
	}

	my @testnames;
	my @cmds;
	push @cmds, "cat $grid_mapfile";
	push @testnames, 'GRID_MAPFILE';
	push @cmds, "ls -l $grid_services";
	push @testnames, 'GRID_SERVICES_LISTING';

	my $a = run_remote_cmds(\@cmds, $opts);

	for ( my $i = 0; $i < scalar(@$a); $i++ )
	{
	    if ( $a->[$i]{'ExitStatus'} == 0 && 
		 $a->[$i]{'RemoteExitStatus'} == 0 )
	    {
		$opts->{'LOCAL_FILES'}{$testnames[$i]} = 
		    stdout_to_file($a->[$i]{'Stdout'}, $opts);
	    }
	}
    }
}


#
# check_grid_mapfile_users
#
# Get a list of the remote user accounts used in the remote grid-mapfile.
#
sub check_grid_mapfile_users($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    if ( ! defined $opts->{'LOCAL_FILES'}{'GRID_MAPFILE'} )
    {
	# Find where the grid-mapfile lives
	my $grid_mapfile = get_grid_mapfile_from_gatekeeper_conf($opts);

	# If we can't find it, guess
	if ( ! defined $grid_mapfile )
	{
	    $grid_mapfile = '/etc/grid-security/grid-mapfile';
	}

	my $f = obtain_remote_file($tests, $opts, $grid_mapfile);

	if ( ! defined $f || ! -f $f ) { return; }

	$opts->{'LOCAL_FILES'}{'GRID_MAPFILE'} = $f;
    }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;

    $opts->{'FILES_HASH'}{'GRID_MAPFILE'} = 
	hashify_grid_mapfile($opts->{'LOCAL_FILES'}{'GRID_MAPFILE'});

    my %rev = reverse %{$opts->{'FILES_HASH'}{'GRID_MAPFILE'}};

    if ( scalar(keys %rev) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = join(",", 
							sort (keys %rev));
    }
    else
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = NONE;
    }
}


#
# check_configured_grid_services
#
# Examine the remote $GLOBUS_LOCATION/etc/grid-services.
#
sub check_configured_grid_services($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $grid_services = get_grid_services_from_gatekeeper_conf();

    # If we can't find it, guess
    if ( ! defined $grid_services )
    {
	$grid_services = $tests->{'RESULTS'}{'globus_location'};
	if ( $grid_services !~ /\/$/ ) { $grid_services .= '/'; }	    
	$grid_services .= 'etc/grid-services';
    }

    if ( ! defined $opts->{'LOCAL_FILES'}{'GRID_SERVICES_LISTING'} )
    {
	my @cmds;
	#push @cmds, "ls -l $grid_services | grep jobmanager | grep -v ~";
	push @cmds, "ls -l $grid_services";

	my $a = run_remote_cmds(\@cmds, $opts);

	if ( scalar(@$a) != 0 )
	{
	    if ( $a->[0]{'ExitStatus'} == 0 && 
		 $a->[0]{'RemoteExitStatus'} == 0 )
	    {
		$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
		$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = NONE;
		$tests->{'STDOUT'}{$opts->{'TESTNAME'}} = $a->[0]{'Stdout'};
		$tests->{'STDERR'}{$opts->{'TESTNAME'}} = $a->[0]{'Stderr'};
	    }
	}
    }
    else
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = NONE;
	$tests->{'STDOUT'}{$opts->{'TESTNAME'}} = 
	    `cat $opts->{'LOCAL_FILES'}{'GRID_SERVICES_LISTING'}`;
	$tests->{'STDERR'}{$opts->{'TESTNAME'}} = undef;
    }

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq SUCCESS )
    {
	my @results = split "\n", $tests->{'STDOUT'}{$opts->{'TESTNAME'}};

	my @jms;

	for ( my $i = 0 ; $i < scalar(@results) ; $i++ )
	{
	    my @tokens = split /\s+/, $results[$i];
	    #if ( scalar(@tokens) == 9 )
	    if ( scalar(@tokens) > 8 )
	    {
		push @jms, $tokens[8];
	    }
	    #elsif ( scalar(@tokens) == 11 )
	    #{
	    #    my ($base, $path) = fileparse($tokens[10]);
	    #    $jmdef = $base;
	    #}
	}

	if ( scalar(@jms) > 0 )
	{
	    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = PASS . "\n  " . 
		join ',', @jms;
	}

	#if ( defined $jmdef )
	#{
	#	if ( defined $tests->{'RESULTS'}{$opts->{'TESTNAME'}} )
	#	{
	#	    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} .= "\n";
	#	}
	#	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} .= 
	#	    "  Default jobmanager is $jmdef";
	#    }

	# Now get a copy of the jobmanagers for later
	@cmds = ();
	foreach my $jm (@jms)
	{
	    push @cmds, "cat $grid_services/$jm;echo";
	}

	$a = run_remote_cmds(\@cmds, $opts);

	for ( my $i = 0; $i < scalar(@$a) ; $i++ )
	{
	    my $f = stdout_to_file($a->[$i]{'Stdout'}, $opts);

	    $opts->{'FILE_HASH'}{$jms[$i]} = hashify_jobmanager($f);

	    unlink $f;
	    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $f\n"; }
	}

	# Now get a copy of the perl modules for later
	my @pms;
	@cmds = ();
	my %h;

	foreach my $key ( sort keys %{$opts->{'FILE_HASH'}} )
	{
	    if ( $key =~ /^jobmanager/ )
	    {
		$h{$key} = undef;
		if ( exists $opts->{'FILE_HASH'}{$key}{'-type'} && 
		     defined $opts->{'FILE_HASH'}{$key}{'-type'} )
		{
		    push @cmds, 'cat ' . $tests->{'RESULTS'}{'globus_location'} . '/lib/perl/Globus/GRAM/JobManager/' . $opts->{'FILE_HASH'}{$key}{'-type'} . '.pm';
		    push @pms, $opts->{'FILE_HASH'}{$key}{'-type'} . '.pm';
		}
	    }
	}

	$a = run_remote_cmds(\@cmds, $opts);

	for ( my $i = 0; $i < scalar(@$a) ; $i++ )
	{
	    $opts->{'LOCAL_FILES'}{$pms[$i]} = 
		stdout_to_file($a->[$i]{'Stdout'}, $opts);
	}
    }
}


#
# check_scheduler_types
#
# Check the scheduler types associated with the remote jobmanagers.
#
sub check_scheduler_types($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    # XXX - FIXME - this is icky
    my %h;

    foreach my $key ( sort keys %{$opts->{'FILE_HASH'}} )
    {
	if ( $key =~ /^jobmanager/ )
	{
	    $h{$key} = undef;
	    if ( exists $opts->{'FILE_HASH'}{$key}{'-type'} && 
		 defined $opts->{'FILE_HASH'}{$key}{'-type'} )
	    {
		$h{$key} = $opts->{'FILE_HASH'}{$key}{'-type'};
	    }
	}
    }

    my $result = SUCCESS;
    foreach my $key ( keys %h )
    {
	if ( ! defined $h{$key} )
	{
	    $result = FAILURE;
	    last;
	}
    }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = $result;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $result . "\n";

    foreach my $key ( sort keys %h )
    {
	my $type = UNKNOWN;
	if ( defined $h{$key} ) { $type = $h{$key}; }

	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} .= 
	    "  $key is of type $type\n";

	$opts->{'JM_TYPES'}{$key} = $type;
    }
}


#
# check_scheduler_paths
#
# Determine the paths to the scheduler commands on the remote site.
#
sub check_scheduler_paths($$)
{
    my ($tests, $opts)  = @_;

    # TODO: almost none of this test is relevant anymore

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    foreach my $key ( keys %{$opts->{'JM_TYPES'}} )
    {
	my $sched = $opts->{'JM_TYPES'}{$key};

	if ( ! defined $sched )  { next; }
	if ( $sched eq UNKNOWN ) { next; }
	if ( $sched eq 'fork' )  { next; }

	if ( ! exists $opts->{'SCHED_PATH'}{$sched} )
	{
	    $opts->{'SCHED_PATH'}{$sched} = undef;
	}

	my %schedpat = ('condor' => '/condor_submit',
			'pbs'    => '/qsub',
			'lsf'    => '/bsub',
			'sge'    => '/qsub',
			);

        # Try default paths
        if ( -e "/usr/bin/$schedpat{$sched}" )
        {
            $opts->{'SCHED_PATH'}{$sched} = "/usr/bin";
            next;
        }

	my $val = undef;

	my @lines;
	if ( exists $opts->{'LOCAL_FILES'}{"$sched.pm"} &&
	     defined $opts->{'LOCAL_FILES'}{"$sched.pm"} &&
	     -f $opts->{'LOCAL_FILES'}{"${sched}.pm"} )
	{
	    my $fh = new IO::File;
	    my $pmfile = $opts->{'LOCAL_FILES'}{"${sched}.pm"};
	    if ( $fh->open("< $pmfile") )
	    {
		while ( my $line = $fh->getline )
		{
		    chomp($line);
		    $line =~ s/^\s+//;
		    $line =~ s/\s+$//;

		    if    ( ! defined $line || length($line) == 0 ) { next;}
		    elsif ( $line =~ m/^\#/ ) { next; }
		    push @lines, $line;
		}
		$fh->close;
	    }
	}

	my @greplist = 
	    grep s/^.*=\s*[\'\"]\s*(.*$schedpat{$sched})\s*[\'\"]\s*;.*$/$1/, @lines;
	if ( scalar(@greplist) > 0 )
	{
	    $val = $greplist[scalar(@greplist)-1];
	}


	if ( defined $val )
	{
	    my ($base, $path) = fileparse($val);
	    if ( defined $path )
	    {
		$path =~ s/\/+$//;
		$opts->{'SCHED_PATH'}{$sched} = $path;
	    }
	}
    }

    # Call it 'SUCCESS' if any non-fork schedulers have PATHs
    my $result = FAILURE;
    foreach my $key ( keys %{$opts->{'JM_TYPES'}} )
    {
	if ( $opts->{'JM_TYPES'}{$key} eq 'fork' ) { next; }

	if ( exists $opts->{'SCHED_PATH'}{$opts->{'JM_TYPES'}{$key}} && 
	     defined $opts->{'SCHED_PATH'}{$opts->{'JM_TYPES'}{$key}} )
	{
	    $result = SUCCESS;
	    last;
	}
    }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = $result;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $result . "\n";

    foreach my $key ( sort keys %{$opts->{'SCHED_PATH'}} )
    {
	my $path = UNKNOWN;
	if ( defined $opts->{'SCHED_PATH'}{$key} )
	{
	    $path = $opts->{'SCHED_PATH'}{$key};
	}

	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} .= 
	    "  Path to $key binaries is $path\n";
    }
}


#
# check_scheduler_status
#
# Check the queue depths
#
sub check_scheduler_status($$)
{
    my ($tests, $opts)  = @_;

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my @cmds;
    my @scheds;

    foreach my $sched ( keys %{$opts->{'SCHED_PATH'}} )
    {
	if ( ! defined $opts->{'SCHED_PATH'}{$sched} ) { next; }
	if ( $opts->{'SCHED_PATH'}{$sched} eq UNKNOWN ) { next; }

	my $cmd = undef;

	if ( $sched eq 'condor' )
	{
	    # XXX FIXME
	    # Any condor binaries need to know where its config file
	    # lives in order for them to work.  This means the condor_config
	    # file lives in in a default place, or the CONDOR_CONFIG
	    # environment variable must be set.  If this is a grid3/OSG
	    # site, this is ONLY set by the top-level setup.sh script (ugh)
	    $cmd = "if [ -f $tests->{'RESULTS'}{'globus_location'}/../setup.sh ]; then . $tests->{'RESULTS'}{'globus_location'}/../setup.sh; fi ; ";
	    $cmd .= $opts->{'SCHED_PATH'}{$sched} . '/condor_q -global';
	}
	elsif ( $sched eq 'pbs' )
	{
	    $cmd = $opts->{'SCHED_PATH'}{$sched} . '/qstat';
	}
	elsif ( $sched eq 'lsf' )
	{
	    $cmd = $opts->{'SCHED_PATH'}{$sched} . '/bjobs -u all';
	}
	elsif ( $sched eq 'sge' )
	{
	    my $a=get_grid3_info_var($tests, $opts, 'OSG_SGE_ROOT');
	    my $b=get_grid3_info_var($tests, $opts, 'OSG_SGE_LOCATION');
        $cmd = "export SGE_ROOT=$a;" . "export PATH=$ENV{'PATH'}:$b/bin;" . 'qstat';
	}

	if ( defined $cmd ) { push @cmds, $cmd; push @scheds, $sched; }
    }

    my $a = run_remote_cmds(\@cmds, $opts);

    if ( scalar(@$a) == 0 ) { return; }

    my $tmp = '';
    for ( my $i = 0 ; $i < scalar(@cmds) ; $i++ )
    {
	my $sched = $scheds[$i];
	$tmp .= "\n  $sched : ";

	my @queue;

	if ( $a->[$i]{'ExitStatus'} == 0 && 
	     $a->[$i]{'RemoteExitStatus'} == 0 )
	{
	    # At least one is good...
	    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
	    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = SUCCESS;

	    # Parse the output
	    my @lines = split /\n/, $a->[$i]{'Stdout'};
	    foreach my $line (@lines)
	    {
		$line =~ s/^\s+//;
		$line =~ s/\s+$//;
		if    ( ! defined $line || length($line) == 0 ) { next;}

		my %job;

		my @tokens = split /\s+/, $line;

		if ( $sched eq 'condor' )
		{
		    if ( $tokens[0] !~ /\d+\.\d+/ ) { next; }
		    if ( scalar(@tokens) < 6 ) { next; }
		    $job{'queue'} = 'default'; # XXX FIXME
		    $job{'id'}    = $tokens[0];
		    $job{'state'} = $tokens[5];
		}
		elsif ( $sched eq 'pbs' )
		{
		    if ( $tokens[0] !~ /\d+\.\S+/ ) { next; }
		    if ( scalar(@tokens) < 6 ) { next; }
		    # Job states:
		    #  R - job is running.
		    #  C - Job is completed and leaving the queue on it's own.
		    #  E - Job is exiting after having run.
		    #  H - Job is held.
		    #  Q - job is queued, eligable to run or routed.
		    #  T - job is being moved to new location.
		    #  I - job is idle.
		    #  W - job is waiting for its execution time
		    #      (-a option) to be reached.
		    #  S - (Unicos only) job is suspend.
		    #  X - Job is removed from the queue
		    $job{'queue'} = $tokens[5];
		    $job{'id'}    = $tokens[0];
		    $job{'state'} = $tokens[4];
		    if    ( $job{'state'} eq 'S' ) { $job{'state'} = 'H'; }
		    elsif ( $job{'state'} eq 'W' ) { $job{'state'} = 'I'; }
		    elsif ( $job{'state'} eq 'Q' ) { $job{'state'} = 'I'; }
		    elsif ( $job{'state'} eq 'T' ) { $job{'state'} = 'I'; }
		    # Skip jobs in uninteresting states
		    elsif ( $job{'state'} eq 'C' ) { next; }
		    elsif ( $job{'state'} eq 'E' ) { next; }
		    elsif ( $job{'state'} eq 'X' ) { next; }
		}
		elsif ( $sched eq 'lsf' )
		{
		    if ( $tokens[0] !~ /\d+/ ) { next; }
		    if ( scalar(@tokens) < 3 ) { next; }
		    # Job states:
		    #  PEND   - Job is pending, not yet started.
		    #  PSUSP  - Job is suspended while pending.
		    #  RUN    - Job is running.
		    #  USUSP  - Job is suspended by the user.
		    #  SSUSP  - Job is suspended by the system.
		    #  UNKWN  - Job is lost.
		    #  WAIT   - Job is waiting for its execution time.
		    #  ZOMBI  - Job is a zombie. (see LSF docs)
		    #  DONE   - Job has terminated with status of 0.
		    #  EXITED - Job has terminated with non-zero status.
		    $job{'queue'} = 'default'; # XXX FIXME
		    $job{'id'}    = $tokens[0];
		    $job{'state'} = $tokens[2];
		    if    ( $job{'state'} eq 'RUN' )   { $job{'state'} = 'R'; }
		    elsif ( $job{'state'} eq 'PEND' )  { $job{'state'} = 'I'; }
		    elsif ( $job{'state'} eq 'PSUSP' ) { $job{'state'} = 'H'; }
		    elsif ( $job{'state'} eq 'SSUSP' ) { $job{'state'} = 'H'; }
		    elsif ( $job{'state'} eq 'USUSP' ) { $job{'state'} = 'H'; }
		    # Skip jobs in uninteresting states
		    elsif ( $job{'state'} eq 'ZOMBI' )     { next; }
		    elsif ( $job{'state'} eq 'DONE' )      { next; }
		    elsif ( $job{'state'} eq 'EXITED' )    { next; }
		    elsif ( $job{'state'} eq 'POST_DONE' ) { next; }
		    elsif ( $job{'state'} eq 'POST_ERR' )  { next; }
		}
		elsif ( $sched eq 'fbsng' )
		{
		    if ( $tokens[0] !~ /\d+/ ) { next; }
		    if ( scalar(@tokens) < 4 ) { next; }
		    # Job states:
		    #  running - Job is running.
		    #  pending - Job is pending.
		    #  ended   - Job is finished.
		    $job{'queue'} = 'default'; # XXX FIXME
		    $job{'id'}    = $tokens[0];
		    $job{'state'} = $tokens[1];
		    if ( $job{'state'} eq 'running' ) { $job{'state'} = 'R'; }
		    elsif ( $job{'state'} eq 'pending' ) { $job{'state'} = 'I'; }
		    # Skip jobs in uninteresting states
		    elsif ( $job{'state'} eq 'ended' ) { next; }
		}
		elsif ( $sched eq 'sge' )
		{
                    if ( $tokens[0] !~ /\d+/ ) { next; }
                    if ( scalar(@tokens) < 6 ) { next; }
                    $job{'queue'} = 'default'; # XXX FIXME
                    $job{'id'}    = $tokens[0];
                    #$job{'state'} = $tokens[2];
                    $job{'state'} = $tokens[4];
                    #print $tokens[4],"\n";
                    # Job states:
                    #  d  - Job is deleted
                    #  t  - Job is transferring.
                    #  r  - Job is running.
                    #  R  - Job is restarted.
                    #  s  - Job is suspended.
                    #  S  - Job is suspended.
                    #  T  - Job is threshold.
                    #  w  - Job is waiting.
                    #  h  - Job is in hold.
                    #  qw,hqw,Rr - these are states that kind of show history.
                    #  but waiting will always show up as qw
                    #  Eqw - means sent back to queing in Error state (fault of SGE)
                    #  so it is actually like held
                    #              The current state is the last character.
                    if    ( $job{'state'} eq 'r' )   { $job{'state'} = 'R'; }
                    #if    ( $job{'state'} eq 'R' )   { $job{'state'} = 'H'; }
                    # the above was commented out as it would make all 
                    # running jobs counted as held
                    if    ( $job{'state'} eq 's' )   { $job{'state'} = 'H'; } 
                    if    ( $job{'state'} eq 'Eqw' )   { $job{'state'} = 'H'; }
                    if    ( $job{'state'} eq 'qw' )   { $job{'state'} = 'I'; }
                    if    ( $job{'state'} eq 'hqw' )   { $job{'state'} = 'H'; }
                    # Skip jobs in uninteresting states
                    elsif ( $job{'state'} eq 'T' ) { next; }
		}

		push @queue, \%job;
	    }

	    my $nr = 0;
	    my $np = 0;
	    my $nh = 0;
	    my $nu = 0;
	    foreach my $job (@queue)
	    {
		if    ( $job->{'state'} eq 'R' ) { $nr++; }
		elsif ( $job->{'state'} eq 'I' ) { $np++; }
		elsif ( $job->{'state'} eq 'H' ) { $nh++; }
		else                             { $nu++; }
	    }
	    $tmp .= "$nr jobs running, $np jobs idle/pending";
	    if ( $nu > 0 ) { $tmp .= ", $nh jobs held/suspended"; }
	    if ( $nu > 0 ) { $tmp .= ", $nu jobs unknown"; }
	}
	else
	{
	    $tmp .= 'NOT WORKING';
	}
    }

    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} .= $tmp
}


# check_osg_version
#
# Find the remote sites osg version by running the program 'osg-version'
# (found in the base directory of the install) and report it
#
sub check_osg_version($$)
{
     my ($tests, $opts)  = @_;
     # This test is optional, really
     $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
 
    if ( check_prereqs($tests, $opts) != 0 ) { return; }
 
     my $d = $tests->{'RESULTS'}{'globus_location'};
 
     if ( ! exists $opts->{'LOCAL_FILES'}{'OSG_VERSION_CHECK'} ||
 	 ! -f $opts->{'LOCAL_FILES'}{'OSG_VERSION_CHECK'} )
     {
 	my @cmds;
 	push @cmds, "if test -x $d/../osg-version ; then OSG_LOCATION=$d/.. $d/../osg-version; else rpm -q osg-ce; fi";
 
 	my $a = run_remote_cmds(\@cmds, $opts);
 
 	if ( scalar(@$a) != 0 )
 	{
 	    if ( $a->[0]{'ExitStatus'} == 0 && 
 		 $a->[0]{'RemoteExitStatus'} == 0 )
 	    {
 		$opts->{'LOCAL_FILES'}{'OSG_VERSION_CHECK'} = 
 		    stdout_to_file($a->[0]{'Stdout'}, $opts);
 	    }
 	}
     }
 
     if ( ! exists $opts->{'LOCAL_FILES'}{'OSG_VERSION_CHECK'} || 
 	 ! -f $opts->{'LOCAL_FILES'}{'OSG_VERSION_CHECK'} )
     {
 	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = NO;
 	return;
     }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    
    my $out = `cat $opts->{'LOCAL_FILES'}{'OSG_VERSION_CHECK'}`;
    chomp($out);
    my $v;
    if (length($out) > 4) {
          if (substr($out, 0, 4) eq "OSG-") { # passed
               $v = substr($out, 4, length($out) - 4);
               $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = YES . '; version ' . $v;
          }
    }
}
#
# check_vdt_version
#
# See if the remote site's middleware deployment is from the VDT
# by looking for (and running, if found) a properly placed vdt-version.info.
#
sub check_vdt_version($$)
{
    my ($tests, $opts)  = @_;

    # This test is optional, really
    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $d = $tests->{'RESULTS'}{'globus_location'};

    if ( ! exists $opts->{'LOCAL_FILES'}{'VDT_VERSION_CHECK'} ||
	 ! -f $opts->{'LOCAL_FILES'}{'VDT_VERSION_CHECK'} )
    {
	my @cmds;
	push @cmds, "if test -x $d/../vdt/bin/vdt-version ; then VDT_LOCATION=$d/.. $d/../vdt/bin/vdt-version; else rpm -q osg-ce; fi";

	my $a = run_remote_cmds(\@cmds, $opts);

	if ( scalar(@$a) != 0 )
	{
	    if ( $a->[0]{'ExitStatus'} == 0 && 
		 $a->[0]{'RemoteExitStatus'} == 0 )
	    {
		$opts->{'LOCAL_FILES'}{'VDT_VERSION_CHECK'} = 
		    stdout_to_file($a->[0]{'Stdout'}, $opts);
	    }
	}
    }

    if ( ! exists $opts->{'LOCAL_FILES'}{'VDT_VERSION_CHECK'} || 
	 ! -f $opts->{'LOCAL_FILES'}{'VDT_VERSION_CHECK'} )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = NO;
	return;
    }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;

    my $out = `cat $opts->{'LOCAL_FILES'}{'VDT_VERSION_CHECK'}`;
    $out =~ s/\n$//;

    if ( $out =~ /VDT\s+\S+\s+(\S+).*/ )
    {
	(my $v = $1) =~ s/\:$//;
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = 
	    YES . '; version ' . $v;
    }
    elsif ( $out =~ /VDT(\d+\.\d+\.\d+).*/ )
    {
	my $v = $1;
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = 
	    YES . '; version ' . $v;
    }
}


#
# check_grid3_info_conf
#
# See if the remote host has a grid3-info.conf in the expected place.
#
sub check_grid3_info_conf($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = NO;

    if ( ! defined $opts->{'LOCAL_FILES'}{'OSG_ATTRIBUTES_CONF'} ) 
    {
        my $rfile = '/etc/osg/osg-attributes.conf';
        if ($tests->{'RESULTS'}{'globus_location'} ne "/" )
        {
	    $rfile = $tests->{'RESULTS'}{'globus_location'} . 
	        '/../monitoring/osg-attributes.conf';
        }

	$opts->{'LOCAL_FILES'}{'OSG_ATTRIBUTES_CONF'} = 
	    obtain_remote_file($tests, $opts, $rfile);
    }

    if ( defined $opts->{'LOCAL_FILES'}{'OSG_ATTRIBUTES_CONF'} )
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = YES;
    }

    if ( defined $opts->{'CACHING_ENABLED'} )
    {
	if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq SUCCESS )
	{
	    my @testnames;
	    my @vars;
	    push @vars, 'GRID3_BASE_DIR';
	    push @testnames, 'grid3_dir';
	    push @vars, 'GRID3_APP_DIR';
	    push @testnames, 'grid3_app_dir';
	    push @vars, 'GRID3_DATA_DIR';
	    push @testnames, 'grid3_data_dir';
	    push @vars, 'GRID3_TMP_DIR';
	    push @testnames, 'grid3_tmp_dir';
	    push @vars, 'GRID3_TMP_WN_DIR';
	    push @testnames, 'grid3_wn_tmp_dir';
	    for ( my $i = 0 ; $i < scalar(@vars) ; $i++ )
	    {
		my $d = get_grid3_info_var($tests, $opts, $vars[$i]);
		$d =~ s/\/+$//;

		if ( defined $d && length($d) > 0 )
		{
		    $tests->{'RESULTS'}{$testnames[$i]} = $d;
		    $tests->{'STATUS'}{$testnames[$i]}  = SUCCESS;
		}
	    }

	    my @cmds;
	    if ( defined  $tests->{'RESULTS'}{'grid3_dir'} )
	    {
		push @cmds, "cat $tests->{'RESULTS'}{'grid3_dir'}/setup.sh";
		$testnames[scalar(@cmds)-1] = 'grid3_setup';

		push @cmds, "cat $tests->{'RESULTS'}{'grid3_dir'}/user-vo-map";
		$testnames[scalar(@cmds)-1] = 'grid3_user_vo_map';
	    }
	    if ( defined  $tests->{'RESULTS'}{'grid3_app_dir'} )
	    {
		push @cmds, "cd $tests->{'RESULTS'}{'grid3_app_dir'}";
		$testnames[scalar(@cmds)-1] = 'grid3_app_dir_exists';

		push @cmds, "dir=$tests->{'RESULTS'}{'grid3_app_dir'} ; pid=`echo \$\$`; touch \\\$dir/.site_verify.\\\$pid && rm -f \\\$dir/.site_verify.\\\$pid";
		$testnames[scalar(@cmds)-1] = 'grid3_app_dir_writable';

		push @cmds, "cat $tests->{'RESULTS'}{'grid3_app_dir'}/etc/grid3-locations.txt";
		$testnames[scalar(@cmds)-1] = 'grid3_site_specific';

		push @cmds, "df -T  $tests->{'RESULTS'}{'grid3_app_dir'}";
		$testnames[scalar(@cmds)-1] = 'grid3_app_dir_df';
	    }
	    if ( defined  $tests->{'RESULTS'}{'grid3_data_dir'} )
	    {
		push @cmds, "cd $tests->{'RESULTS'}{'grid3_data_dir'}";
		$testnames[scalar(@cmds)-1] = 'grid3_data_dir_exists';

		push @cmds, "dir=$tests->{'RESULTS'}{'grid3_data_dir'} ; pid=`echo \$\$`; touch \\\$dir/.site_verify.\\\$pid && rm -f \\\$dir/.site_verify.\\\$pid";
		$testnames[scalar(@cmds)-1] = 'grid3_data_dir_writable';

		push @cmds, "df -T  $tests->{'RESULTS'}{'grid3_data_dir'}";
		$testnames[scalar(@cmds)-1] = 'grid3_data_dir_df';
	    }
	    if ( defined  $tests->{'RESULTS'}{'grid3_tmp_dir'} )
	    {
		push @cmds, "cd $tests->{'RESULTS'}{'grid3_tmp_dir'}";
		$testnames[scalar(@cmds)-1] = 'grid3_tmp_dir_exists';

		push @cmds, "dir=$tests->{'RESULTS'}{'grid3_tmp_dir'} ; pid=`echo \$\$`; touch \\\$dir/.site_verify.\\\$pid && rm -f \\\$dir/.site_verify.\\\$pid";
		$testnames[scalar(@cmds)-1] = 'grid3_tmp_dir_writable';

		push @cmds, "df -T  $tests->{'RESULTS'}{'grid3_tmp_dir'}";
		$testnames[scalar(@cmds)-1] = 'grid3_tmp_dir_df';
	    }

	    my $a = run_remote_cmds(\@cmds, $opts);

	    for ( my $i = 0; $i < scalar(@$a); $i++ )
	    {
		if ( $a->[$i]{'ExitStatus'} == 0 )
		{
		    if ( $testnames[$i] eq 'grid3_site_specific' )
		    {
			$opts->{'LOCAL_FILES'}{'GRID3_LOCATIONS_TXT'} =
			    stdout_to_file($a->[$i]{'Stdout'}, $opts);
		    }
		    elsif ( $testnames[$i] eq 'grid3_setup' )
		    {
			$opts->{'LOCAL_FILES'}{'GRID3_SETUP'} =
			    stdout_to_file($a->[$i]{'Stdout'}, $opts);
		    }
		    elsif ( $testnames[$i] eq 'grid3_user_vo_map' )
		    {
			$opts->{'LOCAL_FILES'}{'GRID3_USER_VO_MAP'} =
			    stdout_to_file($a->[$i]{'Stdout'}, $opts);
		    }
		    elsif ( $testnames[$i] =~ /_df$/ )
		    {
			(my $stem = $testnames[$i]) =~ s/_df$//;
			$opts->{'LOCAL_FILES'}{'DF_OUTPUT'}{$tests->{'RESULTS'}{$stem}} = stdout_to_file($a->[$i]{'Stdout'}, $opts);
		    }
		    else
		    {
			if ( $a->[$i]{'RemoteExitStatus'} == 0 )
			{
			    $tests->{'STATUS'}{$testnames[$i]} = SUCCESS;
			}
			else
			{
			    $tests->{'STATUS'}{$testnames[$i]} = FAILURE;
			}
			$tests->{'STDOUT'}{$testnames[$i]} = $a->[$i]{'Stdout'};
			$tests->{'STDERR'}{$testnames[$i]} = $a->[$i]{'Stderr'};
		    }
		}
	    }
	}
    }
}


#
# check_grid3_user_vo_map
#
# Look for and report on the contents of the grid3-user-vo-map.txt
#
sub check_grid3_user_vo_map($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = NO;

    if ( ! defined $opts->{'LOCAL_FILES'}{'GRID3_USER_VO_MAP'} )
    {
	my @cmds;
	push @cmds, "cat $tests->{'RESULTS'}{'grid3_dir'}/monitoring/grid3-user-vo-map.txt";
	my $a = run_remote_cmds(\@cmds, $opts);
	for ( my $i = 0; $i < scalar(@$a); $i++ )
	{
	    if ( $a->[$i]{'ExitStatus'} == 0 )
	    {
		$opts->{'LOCAL_FILES'}{'GRID3_USER_VO_MAP'} =
		    stdout_to_file($a->[$i]{'Stdout'}, $opts);
	    }
	}
    }

    if ( defined $opts->{'LOCAL_FILES'}{'GRID3_USER_VO_MAP'} )
    {
	my %user_vo_map;
	my $fh = new IO::File;
	if ( $fh->open("< $opts->{'LOCAL_FILES'}{'GRID3_USER_VO_MAP'}") )
	{
	    while ( my $line = $fh->getline )
	    {
		chomp($line);
		$line =~ s/^\s+//;
		$line =~ s/\s+$//;

		if    ( ! defined $line || length($line) == 0 ) { next;}
		elsif ( $line =~ m/^\#/ ) { next; }

		my ($user, $voi) = split /\s+/, $line;
		if ( ! exists $user_vo_map{$voi} )
		{
		    $user_vo_map{$voi} = $user
		}
		else
		{
		    $user_vo_map{$voi} .= ',' . $user
		}
	    }
	    $fh->close;
	}

	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = YES . "\n";
	foreach my $key (keys %user_vo_map)
	{
	    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} .= 
		'  ' . $key . ' users: ' . $user_vo_map{$key} . "\n";
	}
    }
}


#
# check_grid3_setup
#
# See if the remote host has a setup.sh in the expected place.
#
sub check_grid3_setup($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $f = undef;

    if ( ! defined $opts->{'LOCAL_FILES'}{'GRID3_SETUP'} ) 
    {
	my $rfile = $tests->{'RESULTS'}{'grid3_dir'} . '/setup.sh';

	$f = obtain_remote_file($tests, $opts, $rfile);
    }
    else
    {
	$f = $opts->{'LOCAL_FILES'}{'GRID3_SETUP'};
    }

    # Rewrite file without any 'source' or '.' commands
    # so we can source it
    my $fh = new IO::File;

    my ($fhnew, $fnew) = tempfile(
				  TEMPLATE => 'remoteXXXXXX',
				  DIR      => $opts->{'TMPDIR'}, 
				  UNLINK   => 0,
				  );
    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: temporary file $fnew created\n"; }

    if ( $fh->open("< $f") )
    {
	while ( my $line = $fh->getline )
	{
	    chomp($line);
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;
	    if ( ! defined $line || length($line) == 0 ) { next; }
	    if ( $line =~ m/^\#/ ) { next; }
	    if ( $line =~ m/^source/ ) { next; }
	    if ( $line =~ m/^\./ ) { next; }

	    print $fhnew "$line\n";
	}
    }

    close $fhnew;

    close $fh;
    unlink $f;
    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $f\n"; }

    $opts->{'LOCAL_FILES'}{'GRID3_SETUP_SH'} = $fnew;

    if ( defined $opts->{'LOCAL_FILES'}{'GRID3_SETUP_SH'} )
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = YES;
    }
    else
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = NO;
    }
}


#
# check_grid3_site_name
#
# Check grid3-info.conf for the site name.
#
sub check_grid3_site_name($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_SITE_NAME');

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
    else
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    }
}


#
# check_grid3_dir
#
# Check grid3-info.conf for definition of $GRID3.
#
sub check_grid3_dir($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_BASE_DIR');
    $d =~ s/\/+$//;

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
    else
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    }
}


#
# check_osg_grid_dir
#
# Check grid3-info.conf for definition of $OSG_GRID.
#
sub check_osg_grid_dir($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $d = get_grid3_info_var($tests, $opts, 'OSG_GRID');
    $d =~ s/\/+$//;

    if ( defined $d && length($d) > 0 )
    {
        $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
        $tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
    else
    {
        $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    }
}


#
# check_grid3_app_dir
#
# Check grid3-info.conf for definition of $APP.
#
sub check_grid3_app_dir($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_APP_DIR');
    $d =~ s/\/+$//;

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
    else
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    }
}


#
# check_grid3_data_dir
#
# Check grid3-info.conf for definition of $DATA.
#
sub check_grid3_data_dir($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_DATA_DIR');
    $d =~ s/\/+$//;

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
    else
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    }
}


#
# check_grid3_tmp_dir
#
# Check grid3-info.conf for definition of $TMP.
#
sub check_grid3_tmp_dir($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_TMP_DIR');
    $d =~ s/\/+$//;

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
    else
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    }
}


#
# check_grid3_wn_tmp_dir
#
# Check grid3-info.conf for definition of $WN_TMP.
#
sub check_grid3_wn_tmp_dir($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_TMP_WN_DIR');
    $d =~ s/\/+$//;

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
    else
    {
	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
    }
}


#
# check_remote_dir_exists
#
# Check if a remote directory exists.
#
sub check_remote_dir_exists($$$)
{
    my ($tests, $opts, $d)  = @_;

    if ( ! defined $d || length($d) == 0 )
    {
	return undef;
    }

    my $host = undef;
    if ( exists $opts->{'GATEKEEPER_HOST'} )
    {
	$host = $opts->{'GATEKEEPER_HOST'};
    }

    my $port = 2119;
    if ( exists $opts->{'GATEKEEPER_PORT'} )
    {
	$port = $opts->{'GATEKEEPER_PORT'};
    }

    if ( defined $host )
    {
	my $dest = $host;
	if ( defined $port ) { $dest .= ":$port" }

	my $cmd = "globus-job-run $dest/jobmanager-fork /bin/sh -c \"cd $d > /dev/null 2>&1 ; ";
	$cmd .= 'echo \$?"';
	my @f = split "\n", `$cmd`;

	if ( scalar(@f) == 1 )
	{
	    return $f[0];
	}
	else
	{
	    return scalar(@f);
	}
    }

    return -1;
}


#
# check_remote_dir_exists
#
# Check if a remote directory exists.
#
sub check_remote_exec_exists($$$)
{
    my ($tests, $opts, $d)  = @_;

    if ( ! defined $d || length($d) == 0 )
    {
        return undef;
    }

    my $host = undef;
    if ( exists $opts->{'GATEKEEPER_HOST'} )
    {
        $host = $opts->{'GATEKEEPER_HOST'};
    }

    my $port = 2119;
    if ( exists $opts->{'GATEKEEPER_PORT'} )
    {
        $port = $opts->{'GATEKEEPER_PORT'};
    }

    if ( defined $host )
    {
        my $dest = $host;
        if ( defined $port ) { $dest .= ":$port" }

        my $cmd = "globus-job-run $dest/jobmanager-fork /bin/sh -c \"test -x $d > /dev/null 2>&1 ; ";
        $cmd .= 'echo \$?"';
        my @f = split "\n", `$cmd`;

        if ( scalar(@f) == 1 )
        {
            return $f[0];
        }
        else
        {
            return scalar(@f);
        }
    }

return -1;
}



#
# check_grid3_osg_grid_dir_exists
#
# Check for existence of $OSG_GRID.
#
sub check_osg_grid_dir_exists($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
        if ( check_prereqs($tests, $opts) != 0 ) { return; }

        my $d = $tests->{'RESULTS'}{'osg_grid_dir'};

        my $stat = check_remote_dir_exists($tests, $opts, $d);

        if ( ! defined $stat || $stat != 0 )
        {
            $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
            return;
        }

        $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}



#
# check_grid3_app_dir_exists
#
# Check for existence of $APP.
#
sub check_grid3_app_dir_exists($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	my $d = $tests->{'RESULTS'}{'grid3_app_dir'};

	my $stat = check_remote_dir_exists($tests, $opts, $d);

	if ( ! defined $stat || $stat != 0 )
	{
	    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
	    return;
	}

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}


#
# check_grid3_data_dir_exists
#
# Check for existence of $DATA.
#
sub check_grid3_data_dir_exists($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	my $d = $tests->{'RESULTS'}{'grid3_data_dir'};

	my $stat = check_remote_dir_exists($tests, $opts, $d);

	if ( ! defined $stat || $stat != 0 )
	{
	    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
	    return;
	}

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}


#
# check_grid3_tmp_dir_exists
#
# Check for existence of $TMP.
#
sub check_grid3_tmp_dir_exists($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	my $d = $tests->{'RESULTS'}{'grid3_tmp_dir'};

	my $stat = check_remote_dir_exists($tests, $opts, $d);

	if ( ! defined $stat || $stat != 0 )
	{
	    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;
	    return;
	}

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}


#
# check_remote_dir_writable
#
# Check if a remote directory is writable
#
sub check_remote_dir_writable($$$)
{
    my ($tests, $opts, $d)  = @_;

    if ( ! defined $d || length($d) == 0 )
    {
	return undef;
    }

    my $host = undef;
    if ( exists $opts->{'GATEKEEPER_HOST'} )
    {
	$host = $opts->{'GATEKEEPER_HOST'};
    }

    my $port = 2119;
    if ( exists $opts->{'GATEKEEPER_PORT'} )
    {
	$port = $opts->{'GATEKEEPER_PORT'};
    }

    if ( defined $host )
    {
	my $dest = $host;
	if ( defined $port ) { $dest .= ":$port" }

	my $cmd = "globus-job-run $dest/jobmanager-fork /bin/sh -c \"pid=`echo \$\$`; touch $d";
	$cmd .= '/.site_verify.\$pid > /dev/null 2>&1 ; echo \$?';
	$cmd .= '; rm -f $d/.site_verify.\$pid"';
	#print "$cmd\n";
	my @f = split "\n", `$cmd`;

	if ( scalar(@f) == 1 )
	{
	    return $f[0];
	}
	else
	{
	    return scalar(@f);
	}
    }

    return -1;
}


#
# check_grid3_app_dir_writable
#
# Check for writability of $APP.
#
sub check_grid3_app_dir_writable($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

	my $d = $tests->{'RESULTS'}{'grid3_app_dir'};

	my $stat = check_remote_dir_writable($tests, $opts, $d);

	if ( ! defined $stat || $stat != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}


#
# check_grid3_data_dir_writable
#
# Check for writability of $DATA.
#
sub check_grid3_data_dir_writable($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

	my $d = $tests->{'RESULTS'}{'grid3_data_dir'};

	my $stat = check_remote_dir_writable($tests, $opts, $d);

	if ( ! defined $stat || $stat != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}


#
# check_grid3_tmp_dir_writable
#
# Check for writability of $TMP.
#
sub check_grid3_tmp_dir_writable($$)
{
    my ($tests, $opts)  = @_;

    if ( $tests->{'STATUS'}{$opts->{'TESTNAME'}} eq UNTESTED )
    {
	if ( check_prereqs($tests, $opts) != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

	my $d = $tests->{'RESULTS'}{'grid3_tmp_dir'};

	my $stat = check_remote_dir_writable($tests, $opts, $d);

	if ( ! defined $stat || $stat != 0 ) { return; }

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    }
}


#
# check_remote_dir_size
#
# Check remote directory size (in GB)
#
sub check_remote_dir_size($$$)
{
    my ($tests, $opts, $d)  = @_;
    if ( ! defined $d || length($d) == 0 )
    {
	return undef;
    }

    if ( ! exists $opts->{'LOCAL_FILES'}{'DF_OUTPUT'}{$d} || 
	 ! defined $opts->{'LOCAL_FILES'}{'DF_OUTPUT'}{$d} )
    {
	my @cmds;
	push @cmds, "df -T $d";

	my $a = run_remote_cmds(\@cmds, $opts);
	if ( scalar(@a) > 0 && $a->[0]{'ExitStatus'} == 0 )
	{
	    $opts->{'LOCAL_FILES'}{'DF_OUTPUT'}{$d} = 
		stdout_to_file($a->[0]{'Stdout'}, $opts);
	}
    }

    if ( ! exists $opts->{'LOCAL_FILES'}{'DF_OUTPUT'}{$d} || 
	 ! defined $opts->{'LOCAL_FILES'}{'DF_OUTPUT'}{$d} ) { return undef; }

    my @f = split /\n/, `cat $opts->{'LOCAL_FILES'}{'DF_OUTPUT'}{$d}`;

    if ( scalar(@f) < 2 ) { return undef; }

    $f[0] =~ s/Mounted on/Mounted-on/;
    my @labels = split ' ', $f[0];
    shift @f;

    # Be careful to handle cases where df output is more than one line
    my $foo = join ' ', @f;
    my @tokens = split /\s+/, $foo;

    my %df;
    @df{@labels} = @tokens;

    my $size = $df{'Available'};

    # For AFS, we can't use df output
    if ( lc($df{'Type'}) eq 'afs' )
    {
	splice(@f);
        if ( exists $opts->{'GATEKEEPER_HOST'} )
        {
            $host = $opts->{'GATEKEEPER_HOST'};
        }else 
        {
            return undef
        }
        my $dest = $host;
        if ( exists $opts->{'GATEKEEPER_PORT'} )
        {
            $port = $opts->{'GATEKEEPER_PORT'};
            $dest .= ":$port"; 
        }

	$cmd = "globus-job-run $dest/jobmanager-fork /bin/sh -c \"fs examine -path $d | awk '{print \$4}'; ";
	$cmd .= 'echo \$?"';
	@f = split "\n", `$cmd`;

	if ( scalar(@f) > 0 && $f[scalar(@f)-1] == 0 )
	{
	    if ( scalar(@f) > 3 )
	    {
		$f[3] =~ m/(\d+)/;
		$size = $1;
	    }
	}
	else
	{
	    return undef;
	}
    }

    if ( defined $size ) 
    {
	$size /= (1024*1024);
	$size =~ s/^(\d+\.\d\d\d).*$/$1/;
    }
    return $size;
}


#
# check_grid3_app_dir_size
#
# Check size of $APP.
#
sub check_grid3_app_dir_size($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $d = $tests->{'RESULTS'}{'grid3_app_dir'};

    my $stat = check_remote_dir_size($tests, $opts, $d);

    if ( ! defined $stat || $stat < 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $stat . ' GB';
}


#
# check_grid3_data_dir_size
#
# Check size of $DATA.
#
sub check_grid3_data_dir_size($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $d = $tests->{'RESULTS'}{'grid3_data_dir'};

    my $stat = check_remote_dir_size($tests, $opts, $d);

    if ( ! defined $stat || $stat < 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $stat . ' GB';
}


#
# check_grid3_tmp_dir_size
#
# Check size of $TMP.
#
sub check_grid3_tmp_dir_size($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $d = $tests->{'RESULTS'}{'grid3_tmp_dir'};

    my $stat = check_remote_dir_size($tests, $opts, $d);

    if ( ! defined $stat || $stat < 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $stat . ' GB';
}


#
# check_grid3_job_contact
#
# Check grid3-info.conf for definition of $GRID3_JOB_CONTACT.
#
sub check_grid3_job_contact($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_JOB_CONTACT');

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
}


#
# check_grid3_util_contact
#
# Check grid3-info.conf for definition of $GRID3_UTIL_CONTACT.
#
sub check_grid3_util_contact($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_UTIL_CONTACT');

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
}


#
# check_grid3_sponsor_vo
#
# Check grid3-info.conf for definition of $GRID3_SPONSOR.
#
sub check_grid3_sponsor_vo($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $d = get_grid3_info_var($tests, $opts, 'GRID3_SPONSOR');

    if ( defined $d && length($d) > 0 )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = $d;
	$tests->{'STATUS'}{$opts->{'TESTNAME'}}  = SUCCESS;
    }
}


#
# check_grid3_site_specific
#
# Check for site-specific variables at the remote host
#
sub check_grid3_site_specific($$)
{
    my ($tests, $opts)  = @_;

    if ( check_prereqs($tests, $opts) != 0 ) { return; }

    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = FAILURE;

    my $d = $tests->{'RESULTS'}{'grid3_app_dir'};

    if ( ! defined $d || length($d) == 0 ) { return; }

    # This file is optional
    $tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;

    my $f = undef;

    if ( ! defined $opts->{'LOCAL_FILES'}{'GRID3_LOCATIONS_TXT'} )
    {
	$f = obtain_remote_file($tests, $opts, "$d/etc/grid3-locations.txt");
    }
    else
    {
	$f = $opts->{'LOCAL_FILES'}{'GRID3_LOCATIONS_TXT'};
    }

    if ( ! defined $f || ! -f $f )
    {
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = 'NONE';
	return;
    }

    my $fh = new IO::File;

    if ( $fh->open("< $f") )
    {
	# - lines starting with '#' are comments
	# - the first word of the first valid line is the location list name
	# - each following line has the format LogName Version PhName 
	#   where the first 2 are words, the last is the rest)
	my %loclist;
	my $ll = undef;
	while ( my $line = $fh->getline )
	{
	    chomp($line);
	    $line =~ s/^\s+//;
	    $line =~ s/\s+$//;

	    if ( ! defined $line || length($line) == 0 ) { next; }
	    if ( $line =~ m/^\#/ ) { next; }

	    my @tokens = split /\s+/, $line;
	    my $first = $tokens[0];
	    shift @tokens;
	    if ( scalar(@tokens) == 0 )
	    {
		$ll = $first;
		$loclist{$ll} = undef;
	    }
	    elsif ( scalar(@tokens) == 2 )
	    {
		if ( ! defined $ll )
		{
		    $ll = '<No Location List Name>';
		}
		if ( ! exists $loclist{$ll} )
		{
		    $loclist{$ll} = undef;
		}

		$loclist{$ll}{$first}{'Version'} = $tokens[0];
		shift @tokens;
		$loclist{$ll}{$first}{'PhName'} = join ' ', @tokens;
	    }
	    else
	    {
		# malformed - do nothing
	    }
	}

	$tests->{'STATUS'}{$opts->{'TESTNAME'}} = SUCCESS;
	$tests->{'RESULTS'}{$opts->{'TESTNAME'}} = YES;

	foreach my $llname ( keys %loclist )
	{
	    $tests->{'RESULTS'}{$opts->{'TESTNAME'}} .= "\n  $llname\n";

	    foreach my $logname ( sort keys %{$loclist{$llname}} )
	    {
		$tests->{'RESULTS'}{$opts->{'TESTNAME'}} .= 
		    "    " . $logname . " " . 
		    $loclist{$llname}{$logname}{'Version'} . " " . 
		    $loclist{$llname}{$logname}{'PhName'} . "\n";
	    }
	}
    }
}


###################################################################
###################################################################
#
# Begin main
#
###################################################################
###################################################################

# Set up some defaults
my $columns = 79;
if ( defined $ENV{'COLUMNS'} ) { $columns = $ENV{'COLUMNS'} - 1; }
if ( exists $ENV{'GLOBUS_LOCATION'} )
{
    $ENV{'PATH'} = "$ENV{PATH}:$ENV{GLOBUS_LOCATION}/bin";
}
my $localhost = get_fqdn("localhost");
my $user = getlogin();
if ( !defined($user) )
{
    $user = getpwuid($<);
}


# Set up the tests
#
# Tests are performed in a specific order defined in 'ORDER'.
# 'DESCRIP' maps test name to a short description of the test.
# 'SUBS' maps test name to a subroutine.
# 'ENABLED' controls whether or not a test is enabled.
# 'STATUS' records the status of the test (UNTESTED, FAILURE, SUCCESS).
# 'RESULTS' records any test result values.
# 'MESGS' records any messages associated with the test.
# 'PREREQS' describes dependencies among tests.
# 'ACTPRE' maps test name to a subroutine called before a test is performed.
# 'ACTPOST' maps test name to a subroutine called after a test is performed.
# 'ACTONPASS' maps test name to a subroutine called after a passed test.
# 'ACTONFAIL' maps test name to a subroutine called after a failed test.
#
# All the test subroutines have the following prototypes:
#
# void test_subr(reference to tests hash, reference to option hash)
#
my %tests;

$tests{'ORDER'} = [
		   'path', 
		   'proxy', 
		   'remote_host_reachable', 
		   'gatekeeper_service', 
		   'auth',
		   'hello_world',
		   'uptime',
		   'services',
		   'inetd',
		   'globus_location',
		   'hostcert',
		   'gatekeeper_conf',
		   'grid_mapfile_users',
		   'configured_grid_services',
		   'grid3_info_conf',
		   'scheduler_types',
		   'scheduler_paths',
		   'scheduler_status',
		   'vdt_version',
		   'osg_version',
		   'grid3_user_vo_map',
		   'grid3_site_name',
		   'grid3_dir',
		   'osg_grid_dir',
		   'grid3_app_dir',
		   'grid3_data_dir',
		   'grid3_tmp_dir',
		   'grid3_wn_tmp_dir',
		   'osg_grid_dir_exists',
		   'grid3_app_dir_exists',
		   'grid3_data_dir_exists',
		   'grid3_tmp_dir_exists',
		   'grid3_app_dir_writable',
		   'grid3_data_dir_writable',
		   'grid3_tmp_dir_writable',
		   'grid3_app_dir_size',
		   'grid3_data_dir_size',
		   'grid3_tmp_dir_size',
		   'grid3_site_specific',
		   'grid3_job_contact',
		   'grid3_util_contact',
		   'grid3_sponsor_vo',
		   'grid3_setup',
		   'gsiftp_service', 
		   'gsiftp_local_out',
		   'gsiftp_local_in',
		   'gsiftp_local_diffs',
		   ];

$tests{'DESCRIP'}{'path'}  = 'prerequisites needed for testing';
$tests{'DESCRIP'}{'proxy'} = 'for a valid proxy for ' . $user . '@' . $localhost;
$tests{'DESCRIP'}{'remote_host_reachable'} = 'if remote host is reachable';
$tests{'DESCRIP'}{'gatekeeper_service'} = 'for a running gatekeeper';
$tests{'DESCRIP'}{'auth'}               = 'authentication';
$tests{'DESCRIP'}{'hello_world'}        = '\'Hello, World\' application';
$tests{'DESCRIP'}{'uptime'}             = 'remote host uptime';
$tests{'DESCRIP'}{'services'}           = 'remote Internet network services list';
$tests{'DESCRIP'}{'inetd'}              = 'remote Internet servers database configuration';
$tests{'DESCRIP'}{'globus_location'}    = 'for GLOBUS_LOCATION';
$tests{'DESCRIP'}{'hostcert'}           = 'expiration date of remote host certificate';
$tests{'DESCRIP'}{'gatekeeper_conf'}    = 'for gatekeeper configuration file';
$tests{'DESCRIP'}{'gsiftp_service'}     = 'for a running gsiftp server';
$tests{'DESCRIP'}{'gsiftp_local_in'}    = 'gsiftp (local client, remote host -> local host)';
$tests{'DESCRIP'}{'gsiftp_local_out'}   = 'gsiftp (local client, local host -> remote host)';
$tests{'DESCRIP'}{'gsiftp_local_diffs'} = 'that no differences exist between gsiftp\'d files';
$tests{'DESCRIP'}{'grid_mapfile_users'} = 'users in grid-mapfile, if none must be using XACML';
$tests{'DESCRIP'}{'configured_grid_services'} = 'configured grid services';
$tests{'DESCRIP'}{'scheduler_types'}    = 'scheduler types associated with remote jobmanagers';
$tests{'DESCRIP'}{'scheduler_paths'}    = 'for paths to binaries of remote schedulers';
$tests{'DESCRIP'}{'scheduler_status'}   = 'remote scheduler status';
$tests{'DESCRIP'}{'vdt_version'}        = 'if Globus is deployed from the VDT';
$tests{'DESCRIP'}{'osg_version'}        = 'for OSG version';
$tests{'DESCRIP'}{'grid3_info_conf'}    = 'for OSG osg-attributes.conf';
$tests{'DESCRIP'}{'user_vo_map'}  = 'for OSG user-vo-map';
$tests{'DESCRIP'}{'grid3_site_name'}    = 'for OSG site name';
$tests{'DESCRIP'}{'grid3_dir'}          = 'for OSG $GRID3 definition';
$tests{'DESCRIP'}{'osg_grid_dir'} = 'for OSG $OSG_GRID definition';
$tests{'DESCRIP'}{'grid3_app_dir'}      = 'for OSG $APP definition';
$tests{'DESCRIP'}{'grid3_data_dir'}     = 'for OSG $DATA definition';
$tests{'DESCRIP'}{'grid3_tmp_dir'}      = 'for OSG $TMP definition';
$tests{'DESCRIP'}{'grid3_wn_tmp_dir'}   = 'for OSG $WNTMP definition';
$tests{'DESCRIP'}{'osg_grid_dir_exists'}    = 'for OSG $OSG_GRID existence';
$tests{'DESCRIP'}{'grid3_app_dir_exists'}    = 'for OSG $APP existence';
$tests{'DESCRIP'}{'grid3_data_dir_exists'}   = 'for OSG $DATA existence';
$tests{'DESCRIP'}{'grid3_tmp_dir_exists'}    = 'for OSG $TMP existence';
$tests{'DESCRIP'}{'grid3_app_dir_writable'}  = 'for OSG $APP writability';
$tests{'DESCRIP'}{'grid3_data_dir_writable'} = 'for OSG $DATA writability';
$tests{'DESCRIP'}{'grid3_tmp_dir_writable'}  = 'for OSG $TMP writability';
$tests{'DESCRIP'}{'grid3_app_dir_size'}  = 'for OSG $APP available space';
$tests{'DESCRIP'}{'grid3_data_dir_size'} = 'for OSG $DATA available space';
$tests{'DESCRIP'}{'grid3_tmp_dir_size'}  = 'for OSG $TMP available space';
$tests{'DESCRIP'}{'grid3_site_specific'} = 'for OSG additional site-specific variable definitions';
$tests{'DESCRIP'}{'grid3_job_contact'}  = 'for OSG execution jobmanager(s)';
$tests{'DESCRIP'}{'grid3_util_contact'} = 'for OSG utility jobmanager(s)';
$tests{'DESCRIP'}{'grid3_sponsor_vo'}   = 'for OSG sponsoring VO';
$tests{'DESCRIP'}{'grid3_setup'}        = 'for OSG setup.sh';


for ( my $i = 0; $i < scalar(@{$tests{'ORDER'}}); $i++ )
{
    $tests{'SUBS'}{$tests{'ORDER'}[$i]}    = 'check_' . $tests{'ORDER'}[$i];
    if ( ! defined &{$tests{'SUBS'}{$tests{'ORDER'}[$i]}} )
    {
	$tests{'SUBS'}{$tests{'ORDER'}[$i]} = undef;
    }
    $tests{'ENABLED'}{$tests{'ORDER'}[$i]}   = 1;
    $tests{'MESGS'}{$tests{'ORDER'}[$i]}     = undef;
    $tests{'STATUS'}{$tests{'ORDER'}[$i]}    = UNTESTED;
    $tests{'RESULTS'}{$tests{'ORDER'}[$i]}   = undef;
    $tests{'PREREQS'}{$tests{'ORDER'}[$i]}   = undef;
    $tests{'ACTPRE'}{$tests{'ORDER'}[$i]}    = 'pre_check';
    $tests{'ACTPOST'}{$tests{'ORDER'}[$i]}   = 'post_check';
    $tests{'ACTONPASS'}{$tests{'ORDER'}[$i]} = undef;
    $tests{'ACTONFAIL'}{$tests{'ORDER'}[$i]} = undef;
}

$tests{'ACTONFAIL'}{'path'}  = 'print_msgs_and_exit';
$tests{'ACTONFAIL'}{'proxy'} = 'print_msgs_and_exit';

$tests{'PREREQS'}{'gatekeeper_service'} = 'remote_host_reachable';
$tests{'PREREQS'}{'auth'}               = 'gatekeeper_service';
$tests{'PREREQS'}{'hello_world'}        = 'auth';
$tests{'PREREQS'}{'uptime'}             = 'hello_world';
$tests{'PREREQS'}{'services'}           = 'auth';
$tests{'PREREQS'}{'inetd'}              = 'auth';
$tests{'PREREQS'}{'globus_location'}    = 'auth';
$tests{'PREREQS'}{'hostcert'}           = 'globus_location';
$tests{'PREREQS'}{'gatekeeper_conf'}    = 'globus_location';
$tests{'PREREQS'}{'gsiftp_service'}     = 'gatekeeper_conf';
$tests{'PREREQS'}{'gsiftp_local_out'}   = 'gsiftp_service';
$tests{'PREREQS'}{'gsiftp_local_in'}    = 'gsiftp_local_out';
$tests{'PREREQS'}{'grid_mapfile_users'} = 'gatekeeper_conf';
$tests{'PREREQS'}{'configured_grid_services'} = 'gatekeeper_conf';
$tests{'PREREQS'}{'scheduler_types'}    = 'configured_grid_services';
$tests{'PREREQS'}{'scheduler_paths'}    = 'scheduler_types';
$tests{'PREREQS'}{'scheduler_status'}   = 'scheduler_paths';
$tests{'PREREQS'}{'vdt_version'}        = 'globus_location';
$tests{'PREREQS'}{'osg_version'}        = 'globus_location';
$tests{'PREREQS'}{'grid3_info_conf'}    = 'globus_location';
$tests{'PREREQS'}{'grid3_user_vo_map'}  = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_site_name'}    = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_dir'}          = 'grid3_info_conf';
$tests{'PREREQS'}{'osg_grid_dir'} = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_data_dir'}     = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_app_dir'}      = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_tmp_dir'}      = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_wn_tmp_dir'}   = 'grid3_info_conf';
$tests{'PREREQS'}{'osg_grid_dir_exists'}   = 'osg_grid_dir';
$tests{'PREREQS'}{'grid3_data_dir_exists'}   = 'grid3_data_dir';
$tests{'PREREQS'}{'grid3_app_dir_exists'}    = 'grid3_app_dir';
$tests{'PREREQS'}{'grid3_tmp_dir_exists'}    = 'grid3_tmp_dir';
$tests{'PREREQS'}{'grid3_data_dir_writable'} = 'grid3_data_dir_exists';
$tests{'PREREQS'}{'grid3_app_dir_writable'}  = 'grid3_app_dir_exists';
$tests{'PREREQS'}{'grid3_tmp_dir_writable'}  = 'grid3_tmp_dir_exists';
$tests{'PREREQS'}{'grid3_data_dir_size'} = 'grid3_data_dir_exists';
$tests{'PREREQS'}{'grid3_app_dir_size'}  = 'grid3_app_dir_exists';
$tests{'PREREQS'}{'grid3_tmp_dir_size'}  = 'grid3_tmp_dir_exists';
$tests{'PREREQS'}{'grid3_site_specific'} = 'grid3_app_dir_exists';
$tests{'PREREQS'}{'grid3_job_contact'}  = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_util_contact'} = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_sponsor_vo'}   = 'grid3_info_conf';
$tests{'PREREQS'}{'grid3_setup'}        = 'grid3_dir';


# Handle options
my $help         = undef;
my $host         = 'localhost';
my $tmpdir       = '/tmp';
my $disable_cache = undef;
my $show_tests   = undef;
my $skip_auth    = undef;
my $skip_summary = undef;
my $debug        = undef;
my $verbose      = undef;

if ( exists $ENV{'TMPDIR'} )
{
    if ( -d $ENV{'TMPDIR'} )
    {
	$tmpdir = $ENV{'TMPDIR'};
    }
}

GetOptions(
	   "help"         => \$help,
	   "host=s"       => \$host,
	   "tmpdir=s"     => \$tmpdir,
	   "disable-caching" => \$disable_cache,
	   "show-tests"   => \$show_tests,
	   "debug"        => \$debug,
	   "verbose"      => \$verbose,
	   );

if ( defined $help )
{
    usage();
    exit;
}

if ( defined $show_tests )
{
    show_tests();
    exit;
}

my @sites;
if ( defined $host )
{
    if ( $host !~ m/,$/ ) { $host .= ","; }

    my @foo = split ",", $host;

    for ( my $i=(scalar(@foo)-1); $i >= 0; $i-- )
    {
	# remove empty elements
	if ( length($foo[$i]) == 0 )
	{
	    splice(@foo, $i, 1);
	    next;
	}

	# lowercase the names
	$foo[$i] = lc $foo[$i];

	my $h = $foo[$i];
	$h =~ s/:.*$//;
	my $p = $foo[$i];
	if ( $p =~ m/^.*:/ )
	{
	    $p =~ s/^.*://;
	}
	else
	{
	    $p = 2119;
	}

	my %hash;
	$hash{'GATEKEEPER_HOST'} = $h;
	$hash{'GATEKEEPER_PORT'} = $p;
	if ($hash{'GATEKEEPER_HOST'} eq "localhost")
	{
	    $hash{'GATEKEEPER_HOST'} = get_fqdn($hash{'GATEKEEPER_HOST'});
	}

	push @sites, {%hash};
    }
}

# Here we go.
print get_separator("", "=", $columns) . "\n";
print "Info: Site verification initiated at " . gmtime() . " GMT.\n";
print get_separator("", "=", $columns) . "\n";

# Run the tests for each remote host.
my %opts;
$opts{'DEBUG'}     = $debug;
$opts{'VERBOSE'}   = $verbose;
$opts{'TMPDIR'}    = $tmpdir;
$opts{'COLUMNS'}   = $columns;
$opts{'USER'}      = $user;
$opts{'LOCALHOST'} = $localhost;
if ( ! defined $disable_cache )
{
    $opts{'CACHING_ENABLED'} = 1;
}

for ( my $j = 0; $j < scalar(@sites); $j++ )
{
    if ( exists $sites[$j]->{'GATEKEEPER_HOST'} )
    {
	$opts{'GATEKEEPER_HOST'} = $sites[$j]->{'GATEKEEPER_HOST'};
    }

    if ( exists $sites[$j]->{'GATEKEEPER_PORT'} )
    {
	$opts{'GATEKEEPER_PORT'} = $sites[$j]->{'GATEKEEPER_PORT'};
    }

    # Banner
    my $title = "Begin " . $opts{'GATEKEEPER_HOST'} . " at " . 
	gmtime() . " GMT";
    print get_separator("", "-", $columns) . "\n";
    print get_separator($title, "-", $columns) . "\n";
    print get_separator("", "-", $columns) . "\n";

    for ( my $i = 0; $i < scalar(@{$tests{'ORDER'}}); $i++ )
    {
	my $t = $tests{'ORDER'}[$i];

	$opts{'TESTNAME'} = $t;

	if ( ! defined $tests{'SUBS'}{$t} )
	{
	    print "Info: No implementation for test \"$t\"; skipping...\n";
	    next;
	}

	if ( ! defined $tests{'ENABLED'}{$t} )
	{
	    print "Warn: Test \"$t\" is disabled; skipping... \n";
	    next;
	}

	# Pre-testing hook
	if ( defined $tests{'ACTPRE'}{$t} )
	{
	    if ( defined &{$tests{'ACTPRE'}{$t}} )
	    {
		&{$tests{'ACTPRE'}{$t}}(\%tests, \%opts);
	    }
	    else
	    {
		print "Warn: No pre-testing subroutine \"$tests{ACTPRE}{$t}\" exists (test \"$t\")\n";
	    }
	}

	# Perform the test
	&{$tests{'SUBS'}{$t}}(\%tests, \%opts);

	# Post-testing hook
	if ( defined $tests{'ACTPOST'}{$t} )
	{
	    if ( defined &{$tests{'ACTPOST'}{$t}} )
	    {
		&{$tests{'ACTPOST'}{$t}}(\%tests, \%opts);
	    }
	    else
	    {
		print "Warn: No post-testing subroutine \"$tests{ACTPOST}{$t}\" exists (test \"$t\")\n";
	    }
	}
    }

    # Banner
    $title = "End " . $opts{'GATEKEEPER_HOST'} . " at " . gmtime() . " GMT";
    print get_separator("", "-", $columns) . "\n";
    print get_separator($title, "-", $columns) . "\n";
    print get_separator("", "-", $columns) . "\n";

    # Cleanup
    delete $opts{'GATEKEEPER_HOST'};
    delete $opts{'GATEKEEPER_PORT'};
    if ( defined $opts{'GSIFTP_HOST'} ) { delete $opts{'GSIFTP_HOST'}; }
    if ( defined $opts{'GSIFTP_PORT'} ) { delete $opts{'GSIFTP_PORT'}; }
    delete $tests{'RESULTS'};
    delete $tests{'STATUS'};
    delete $tests{'MESGS'};
    foreach my $f ( keys %{$opts{'LOCAL_FILES'}{'DF_OUTPUT'}} )
    {
	if ( -f $opts{'LOCAL_FILES'}{'DF_OUTPUT'}{$f} )
	{
	    unlink $opts{'LOCAL_FILES'}{'DF_OUTPUT'}{$f};
	    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $opts{'LOCAL_FILES'}{'DF_OUTPUT'}{$f}\n"; }
	}
	delete $opts{'LOCAL_FILES'}{'DF_OUTPUT'}{$f};
    }
    delete $opts{'LOCAL_FILES'}{'DF_OUTPUT'};
    foreach my $f ( keys %{$opts{'LOCAL_FILES'}} )
    {
	if ( -f $opts{'LOCAL_FILES'}{$f} )
	{
	    unlink $opts{'LOCAL_FILES'}{$f};
	    if ( defined $opts->{'DEBUG'} ) { print "DEBUG: Removing file $opts{'LOCAL_FILES'}{$f}\n"; }
	}
	delete $opts{'LOCAL_FILES'}{$f};
    }
    delete $opts{'LOCAL_FILES'};
}

# We are done.
print get_separator("", "=", $columns) . "\n";
print "Info: Site verification completed at " . gmtime() . " GMT.\n";

#
# show_tests()
#
# Display the names of tests to be run, along with a short description.
#
sub show_tests {
    my $longest = 0;
    foreach my $t (@{$tests{'ORDER'}})
    {
	if ( length($t) > $longest )
	{
	    $longest = length($t);
	}
    }

    foreach my $t (@{$tests{'ORDER'}})
    {
	my $s = sprintf "%-${longest}s", $t;
	print "$s $tests{'DESCRIP'}{$t}\n";
    }
}


#
# usage()
#
# Print the usage for this script.
#
sub usage {
    my $scriptname = fileparse("$0");
    print "\n";
    print "Usage: $scriptname [OPTIONS]\n";
    print "\n";
    print "Perfom site verification procedures on a grid computing resource\n";
    print "and display test results.\n";
    print "\n";
    print "Options:\n";
    print "\n";
    print "  --help            Display this message and exit.\n";
    print "  --disable-caching Disable the feature that allows running of multiple\n";
    print "                    tests per authentication.  This is done by default to\n";
    print "                    speed up the verification.\n";
    print "  --debug           Turn on some debugging messages.\n";
    print "  --host=...        Only query selected host(s).  Multiple hosts\n";
    print "                    can be specified in a comma-separated list.\n";
    print "                    Can query non-official sites in this manner.\n";
    print "  --show-tests      Display the tests to be run along with a short\n";
    print "                    description.\n";
    print "  --tmpdir=<dir>    Use <dir> for writing of temporary files.\n";
#    print "  --verbose         Verbose output from authentication checks.\n";
    print "\n";
}


#
# get_fmts()
#
# Return perl formats for a table.
#
sub get_fmts {
    my ($num, $stem, $sites_ref, $keys_ref, $descrip) = @_;

    my $header_fmt = "format HEADER$num = \n";
    $header_fmt .= "Table $num: $descrip\n\n";
    my $data_fmt = "format DATA$num = \n";

    my $betw_field_spacing = "  ";

    my $underscore;
    my @vars;

    for (my $i = 0; $i < scalar(@$keys_ref); $i++) {
	my $longest = 0;
	foreach my $site (@$sites_ref) {
	    if (defined($site->{"$stem"})) {
		my $key = $keys_ref->[$i];
		$key =~ s/\$/\\\$/g;
		if (defined($site->{"$stem"}->{"$key"})) {
		    if (length($site->{"$stem"}->{"$key"}) > $longest) { 
			$longest = length($site->{"$stem"}->{"$key"});
		    }
		}
	    }
	}

	if ($longest < length($keys_ref->[$i])) { 
	    $longest = length($keys_ref->[$i]);
	}

	my $j = 0;

	while ($j < $longest) {
	    if ($j == 0) {
		$header_fmt .= $keys_ref->[$i];
	    } elsif ($j >= length($keys_ref->[$i])) {
		$header_fmt .= " ";
	    }
	    if ($j == 0) {
		$data_fmt .= "@";
	    } else {
		$data_fmt .= "<";
	    }
	    $underscore .= "-";
	    $j++;
	}

	$j = 0;
	while ($j < length($betw_field_spacing)) {
	    $header_fmt .= " ";
	    $data_fmt   .= " ";
	    $underscore .= "-";
	    $j++;
	}

	my $key = $keys_ref->[$i];
	$key =~ s/\$/\\\$/g;
	push @vars, "\$site->{\"$stem\"}->{\"$key\"}"
    }

    $header_fmt .= "\n" . $underscore . "\n.\n";
    $data_fmt   .= "\n" . join(",",@vars) . "\n.\n";

    return ($header_fmt, $data_fmt);
}

#
# generate_summary_tables()
#
# Print summary info in tabular fashion.
#
sub generate_summary_tables {

    my @sites = @_;

    if (scalar(@sites) > 0) {
	print get_separator("", "=", $columns) . "\n";

	print get_separator("", "-", $columns) . "\n";
	print get_separator("Begin Site Verification Summary Tables", "-", $columns) . "\n";
	print get_separator("", "-", $columns) . "\n\n";

	my @keys;
	my %hash;
	$hash{"Key"} = "Mds";
	$hash{"Mds"} = ["Hostname", "SiteName", "VOsponsor", 
			"SchemaOK", "MDSVersion"];
	$hash{"Description"} = "MDS-advertised/derived info";
	push @keys, {%hash};

	$hash{"Key"} = "Mds";
	$hash{"Mds"} = ["Hostname", "Schedulers", "PolicyURL"];
	$hash{"Description"} = "MDS-advertised/derived info (cont.)";
	push @keys, {%hash};

	$hash{"Key"} = "Mds";
	$hash{"Mds"} = ["Hostname", "\$GRID3", "\$APP", "\$DATA", "\$TMP", 
			"\$WNTMP"];
	$hash{"Description"} = "MDS-advertised/derived info (cont.)";
	push @keys, {%hash};

	$hash{"Key"} = "Auth";
	$hash{"Auth"} = ["Hostname", "Authenticate", "gsiftpTo", "gsiftpFrom", 
			 "HelloWorld"];
	$hash{"Description"} = "Checks requiring authentication";
	push @keys, {%hash};

	$hash{"Key"} = "Auth";
	$hash{"Auth"} = ["Hostname", "MultiVOgrid-mapfile"];
	$hash{"Description"} = "Checks requiring authentication (cont.)";
	push @keys, {%hash};

	$hash{"Key"} = "Auth";
	$hash{"Auth"} = ["Hostname", "\$APPexists", "\$APPhasSpace", 
			 "\$APPwritable"];
	$hash{"Description"} = "Checks requiring authentication & MDS (\$APP)";
	push @keys, {%hash};

	$hash{"Key"} = "Auth";
	$hash{"Auth"} = ["Hostname", "\$DATAexists", "\$DATAhasSpace", 
			 "\$DATAwritable"];
	$hash{"Description"} = "Checks requiring authentication & MDS (\$DATA)";
	push @keys, {%hash};

	$hash{"Key"} = "Auth";
	$hash{"Auth"} = ["Hostname", "\$TMPexists", "\$TMPhasSpace", 
			 "\$TMPwritable"];
	$hash{"Description"} = "Checks requiring authentication & MDS (\$TMP)";
	push @keys, {%hash};

	$hash{"Key"} = "Auth";
	$hash{"Auth"} = ["Hostname", "\$WNTMPexists", "\$WNTMPhasSpace", 
			 "\$WNTMPwritable"];
	$hash{"Description"} = "Checks requiring authentication & MDS (\$WNTMP)";
	push @keys, {%hash};

	for (my $i = 0; $i < scalar(@keys); $i++) {
	    my ($fmt_header, $fmt_data) = 
		get_fmts($i, $keys[$i]->{"Key"}, \@sites, 
			 $keys[$i]->{"$keys[$i]->{\"Key\"}"},
			 $keys[$i]->{"Description"});

	    #print "\n$fmt_header\n";
	    #print "\n$fmt_data\n";

	    eval $fmt_header;
	    eval $fmt_data;

	    $^L = "\n\n";
	    $^ = "HEADER$i";
	    $~ = "DATA$i";

	    for (my $j = 0; $j < scalar(@sites); $j++) {
		# must be global...
		$site = $sites[$j];
		write;
	    }

	    $- = 0;
	}

	print "\n";
	print get_separator("", "-", $columns) . "\n";
	print get_separator("End Site Verification Summary Tables", "-", $columns) . "\n";
	print get_separator("", "-", $columns) . "\n";
    }
}

########## Utility/Debug routines ##############

#
# print_array_of_hash_refs()
#
# A debugging routine - print the contents of the of an array of 
# hash references.
#
sub print_array_of_hash_refs {

    foreach my $foo (@_) {
	print "********\n";
	foreach my $key (keys(%{$foo})) {
	    print "\t$key = $foo->{$key}\n";
	}
    }    
}


#
# get_separator()
#
# Return a separator string composed of a centered title surrounded by a 
# single character extended to length "len".
#
sub get_separator {
    my ($title, $char, $len, $just) = @_;

    if (!defined($len)) { $len = 79; }

    my $separator = "";
    my $tlen = length($title);

    if ($tlen >= $len) { return $title; }

    if ($tlen != 0) {
	if (defined($just)) {
	    if (lc($just) eq "l") {
		$title .= " ";
	    } elsif (lc($just) eq "r") {
		$title = " " . $title;
	    } else {
		$title = " " . $title . " ";
	    }
	} else {
	    $title = " " . $title . " ";
	}
	$tlen = length($title);
    }

    if ($tlen >= $len) { return $title; }

    my $before = ($len - $tlen)/2;
    if (defined($just)) {
	if (lc($just) eq "l") {
	    $before = 0;
	} elsif (lc($just) eq "r") {
	    $before = $len - $tlen;
	}
    }

    while (length($separator) < $before) {
	$separator .= $char;
    }

    $separator .= $title;

    while (length($separator) < $len) {
	$separator .= $char;
    }

    return $separator;
}


